import itertools
import unittest
import math
import os
import shutil
import tempfile

import numpy as np
from numpy.testing import assert_almost_equal
import scipy
from io import StringIO

from packaging.version import Version

try:
    from parameterized import parameterized
except ImportError:
    from openmdao.utils.assert_utils import SkipParameterized as parameterized

import openmdao.api as om
from openmdao.components.exec_comp import _expr_dict, _temporary_expr_dict
from openmdao.utils.assert_utils import assert_near_equal, assert_check_partials, assert_warning
from openmdao.utils.general_utils import env_truthy
from openmdao.utils.testing_utils import force_check_partials
from openmdao.utils.om_warnings import OMDeprecationWarning, SetupWarning

_ufunc_test_data = {
    'min': {
        'str': 'f=min(x)',
        'check_func': np.min,
        'args': { 'f': {'val': np.zeros(6)},
                  'x': {'val': np.random.random(6)}}},
    'max': {
        'str': 'f=max(x)',
        'check_func': np.max,
        'args': { 'f': {'val': np.zeros(6)},
                  'x': {'val': np.random.random(6)}}},
    'diff': {
        'str': 'f=diff(x)',
        'check_func': np.diff,
        'args': { 'f': {'val': np.zeros(5)},
                  'x': {'val': np.random.random(6)}}},
    'abs': {
        'str': 'f=abs(x)',
        'check_func': np.abs,
        'args': { 'f': {'val': np.zeros(6)},
                  'x': {'val': np.random.random(6)}}},
    'acos': {
        'str': 'f=acos(x)',
        'check_func': np.arccos,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6) - 0.5}}},
    'arccos': {
        'str': 'f=arccos(x)',
        'check_func': np.arccos,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6) - 0.5}}},
    'arccosh': {
        'str': 'f=arccosh(x)',
        'check_func': np.arccosh,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': 1.1 + np.random.random(6)}}},
    'acosh': {
        'str': 'f=acosh(x)',
        'check_func': np.arccosh,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': 1.1 + np.random.random(6)}}},
    'arange': {
        'str': 'f=arange(0,10,2)',
        'check_val': np.arange(0, 10, 2),
        'args': {'f': {'val': np.zeros(5)}}},
    'arcsin': {
        'str': 'f=arcsin(x)',
        'check_func': np.arcsin,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6) - .5}}},
    'arcsinh': {
        'str': 'f=arcsinh(x)',
        'check_func': np.arcsinh,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6)}}},
    'asinh': {
        'str': 'f=asinh(x)',
        'check_func': np.arcsinh,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6)}}},
    'asin': {
        'str': 'f=asin(x)',
        'check_func': np.arcsin,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6) - .5}}},
    'arctan': {
        'str': 'f=arctan(x)',
        'check_func': np.arctan,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6)}}},
    'arctan2': {
        'str': 'f=arctan2(y, x)',
        'check_val': np.array([-2.35619449, -0.78539816,  0.78539816,  2.35619449]),
        'args': {'f': {'val': np.zeros(4)},
                 'x': {'val': np.array([-1, +1, +1, -1])},
                 'y': {'val': np.array([-1, -1, +1, +1])}}},
    'atan': {
        'str': 'f=atan(x)',
        'check_func': np.arctan,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6)}}},
    'cos': {
        'str': 'f=cos(x)',
        'check_func': np.cos,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6)}}},
    'cosh': {
        'str': 'f=cosh(x)',
        'check_func': np.cosh,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6)}}},
    'dot': {
        'str': 'f=dot(x, y)',
        'check_func': np.dot,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6)},
                 'y': {'val': np.random.random(6)}}},
    'e': {
        'str': 'f=e',
        'check_val': np.e,
        'args': {'f': {'val': 0.0}}},
    'erf': {
        'str': 'f=erf(x)',
        'check_func': scipy.special.erf,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6)}}},
    'erfc': {
        'str': 'f=erfc(x)',
        'check_func': scipy.special.erfc,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6)}}},
    'exp': {
        'str': 'f=exp(x)',
        'check_func': np.exp,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6)}}},
    'expm1': {
        'str': 'f=expm1(x)',
        'check_func': np.expm1,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6)}}},
    'fmax': {
        'str': 'f=fmax(x, y)',
        'check_func': np.fmax,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6)},
                 'y': {'val': np.random.random(6)}}},
    'fmin': {
        'str': 'f=fmin(x, y)',
        'check_func': np.fmin,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6)},
                 'y': {'val': np.random.random(6)}}},
    'inner': {
        'str': 'f=inner(x, y)',
        'check_func': np.inner,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6)},
                 'y': {'val': np.random.random(6)}}},
    'isinf': {
        'str': 'f=isinf(x)',
        'check_func': np.isinf,
        'args': {'f': {'val': np.zeros(3)},
                 'x': {'val': [0, np.inf, 5.0]}}},
    'isnan': {
        'str': 'f=isnan(x)',
        'check_func': np.isnan,
        'args': {'f': {'val': np.zeros(3)},
                 'x': {'val': [0, np.nan, np.nan]}}},
    'kron': {
        'str': 'f=kron(x, y)',
        'check_func': np.kron,
        'args': {'f': {'val': np.zeros(36)},
                 'x': {'val': np.random.random(6)},
                 'y': {'val': np.random.random(6)}}},
    'linspace': {
        'str': 'f=linspace(0,10,50)',
        'check_val': np.linspace(0, 10, 50),
        'args': {'f': {'val': np.zeros(50)}}},
    'log': {
        'str': 'f=log(x)',
        'check_func': np.log,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6) + 0.1}}},
    'log10': {
        'str': 'f=log10(x)',
        'check_func': np.log10,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6) + 0.1}}},
    'log1p': {
        'str': 'f=log1p(x)',
        'check_func': np.log1p,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6)}}},
    'matmul': {
        'str': 'f=matmul(x, y)',
        'check_func': np.matmul,
        'args': {'f': {'val': np.zeros((3, 1))},
                 'x': {'val': np.random.random((3, 3))},
                 'y': {'val': np.random.random((3, 1))}}},
    'maximum': {
        'str': 'f=maximum(x, y)',
        'check_func': np.maximum,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6)},
                 'y': {'val': np.random.random(6)}}},
    'minimum': {
        'str': 'f=minimum(x, y)',
        'check_func': np.minimum,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6)},
                 'y': {'val': np.random.random(6)}}},
    'ones': {
        'str': 'f=ones(21)',
        'check_val': np.ones(21),
         'args': {'f': {'val': np.zeros(21)}}},
    'outer': {
        'str': 'f=outer(x, y)',
        'check_func': np.outer,
        'args': {'f': {'val': np.zeros((6, 6))},
                 'x': {'val': np.random.random(6)},
                 'y': {'val': np.random.random(6)}}},
    'pi': {
        'str': 'f=pi',
        'check_val': np.pi,
        'args': {'f': {'val': 0.0}}},
    'power': {
        'str': 'f=power(x, y)',
        'check_func': np.power,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6)},
                 'y': {'val': np.random.random(6) + 1.0}}},
    'prod': {
        'str': 'f=prod(x)',
        'check_func': np.prod,
        'args': {'f': {'val': 0.0},
                 'x': {'val': np.random.random(6)}}},
    'sin': {
        'str': 'f=sin(x)',
        'check_func': np.sin,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6)}}},
    'sinh': {
        'str': 'f=sinh(x)',
        'check_func': np.sinh,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6)}}},
    'sum': {
        'str': 'f=sum(x)',
        'check_func': np.sum,
        'args': {'f': {'val': 0.0},
                 'x': {'val': np.random.random(6)}}},
    'tan': {
        'str': 'f=tan(x)',
        'check_func': np.tan,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6)}}},
    'tanh': {
        'str': 'f=tanh(x)',
        'check_func': np.tanh,
        'args': {'f': {'val': np.zeros(6)},
                 'x': {'val': np.random.random(6)}}},
    'tensordot': {
        'str': 'f=tensordot(x, y)',
        'check_func': np.tensordot,
        'args': {'f': {'val': 0.0},
                 'x': {'val': np.random.random((6, 6))},
                 'y': {'val': np.random.random((6, 6))}}},
    'zeros': {
        'str': 'f=zeros(21)',
        'check_val': np.zeros(21),
        'args': {'f': {'val': np.zeros(21)}}},
}


class TestExecComp(unittest.TestCase):

    def test_missing_partial_warn(self):
        p = om.Problem()
        model = p.model
        comp = om.ExecComp('z=3.0*x + 2.5*y')
        model.add_subsystem('comp', comp)
        comp.declare_partials('z', 'x', method='fd')
        p.setup()
        with assert_warning(UserWarning, "'comp' <class ExecComp>: The following partial derivatives have not been declared so they are assumed to be zero: ['z' wrt 'y']."):
            p.final_setup()

    def test_colon_vars(self):
        prob = om.Problem()
        prob.model.add_subsystem('C1', om.ExecComp('y=foo:bar+1.'))
        with self.assertRaises(Exception) as context:
            prob.setup()
        self.assertEqual(str(context.exception),
                         "'C1' <class ExecComp>: failed to compile expression 'y=foo:bar+1.'.")

    def test_bad_kwargs(self):
        prob = om.Problem()
        prob.model.add_subsystem('C1', om.ExecComp('y=x+1.', xx=2.0))
        with self.assertRaises(Exception) as context:
            prob.setup()
        self.assertEqual(str(context.exception),
                         "'C1' <class ExecComp>: arg 'xx' in call to ExecComp() does not refer to any variable "
                         "in the expressions ['y=x+1.']")

    def test_bad_kwargs_meta(self):
        prob = om.Problem()
        prob.model.add_subsystem('C1', om.ExecComp('y=x+1.',
                                                   x={'val': 2, 'low': 0, 'high': 10, 'units': 'ft'}))
        with self.assertRaises(Exception) as context:
            prob.setup()
        self.assertEqual(str(context.exception),
                         "'C1' <class ExecComp>: the following metadata names were not recognized for "
                         "variable 'x': ['high', 'low']")

    def test_name_collision_const(self):
        prob = om.Problem()
        prob.model.add_subsystem('C1', om.ExecComp('e=x+1.'))
        with self.assertRaises(Exception) as context:
            prob.setup()
        self.assertEqual(str(context.exception),
                         "'C1' <class ExecComp>: cannot assign to variable 'e' because it's already defined "
                         "as an internal function or constant.")

    def test_name_collision_func(self):
        prob = om.Problem()
        prob.model.add_subsystem('C1', om.ExecComp('sin=x+1.'))
        with self.assertRaises(Exception) as context:
            prob.setup()
        self.assertEqual(str(context.exception),
                         "'C1' <class ExecComp>: cannot assign to variable 'sin' because it's already defined "
                         "as an internal function or constant.")

    def test_func_as_rhs_var(self):
        prob = om.Problem()
        prob.model.add_subsystem('C1', om.ExecComp('y=sin+1.'))
        with self.assertRaises(Exception) as context:
            prob.setup()
        self.assertEqual(str(context.exception),
                         "'C1' <class ExecComp>: cannot use 'sin' as a variable because it's already defined "
                         "as an internal function or constant.")

    def test_mixed_type(self):
        prob = om.Problem()
        C1 = prob.model.add_subsystem('C1', om.ExecComp('y=sum(x)',
                                                        x=np.arange(10, dtype=float)))
        prob.setup()

        # Conclude setup but don't run model.
        prob.final_setup()

        self.assertTrue('x' in C1._inputs)
        self.assertTrue('y' in C1._outputs)

        prob.set_solver_print(level=0)
        prob.run_model()

        assert_near_equal(C1._outputs['y'], 45.0, 0.00001)

    def test_simple(self):
        prob = om.Problem()
        C1 = prob.model.add_subsystem('C1', om.ExecComp('y=x+1.', x=2.0))

        prob.setup()

        # Conclude setup but don't run model.
        prob.final_setup()

        self.assertTrue('x' in C1._inputs)
        self.assertTrue('y' in C1._outputs)

        prob.set_solver_print(level=0)
        prob.run_model()

        assert_near_equal(C1._outputs['y'], 3.0, 0.00001)

    def test_for_spaces(self):
        prob = om.Problem()
        C1 = prob.model.add_subsystem('C1', om.ExecComp('y = pi * x', x=2.0))

        prob.setup()

        # Conclude setup but don't run model.
        prob.final_setup()

        self.assertTrue('x' in C1._inputs)
        self.assertTrue('y' in C1._outputs)
        self.assertTrue('pi' not in C1._inputs)

        prob.set_solver_print(level=0)
        prob.run_model()

        assert_near_equal(C1._outputs['y'], 2 * math.pi, 0.00001)

    def test_units(self):
        prob = om.Problem()
        prob.model.add_subsystem('indep', om.IndepVarComp('x', 100.0, units='cm'))
        C1 = prob.model.add_subsystem('C1', om.ExecComp('y=x+z+1.',
                                                        x={'val': 2.0, 'units': 'm'},
                                                        y={'units': 'm'},
                                                        z=2.0))
        prob.model.connect('indep.x', 'C1.x')

        prob.setup()

        prob.set_solver_print(level=0)
        prob.run_model()

        assert_near_equal(C1._outputs['y'], 4.0, 0.00001)

    def test_units_varname(self):
        prob = om.Problem()

        with self.assertRaises(TypeError) as cm:
            prob.model.add_subsystem('C1', om.ExecComp('y=x+units+1.',
                                                       x={'val': 2.0, 'units': 'm'},
                                                       y={'units': 'm'},
                                                       units=2.0))

        self.assertEqual(str(cm.exception),
                         "ExecComp: Value (2.0) of option 'units' has type 'float', "
                         "but type 'str' was expected.")

    def test_units_varname_str(self):
        prob = om.Problem()

        with self.assertRaises(ValueError) as cm:
            prob.model.add_subsystem('C1', om.ExecComp('y=x+units+1.',
                                                       x={'val': 2.0, 'units': 'm'},
                                                       y={'units': 'm'},
                                                       units='two'))

        self.assertEqual(str(cm.exception), "The units 'two' are invalid.")

    def test_units_varname_novalue(self):
        prob = om.Problem()
        prob.model.add_subsystem('indep', om.IndepVarComp('x', 100.0, units='cm'))
        C1 = prob.model.add_subsystem('C1', om.ExecComp('y=x+units+1.',
                                                        x={'val': 2.0, 'units': 'm'},
                                                        y={'units': 'm'}))
        prob.model.connect('indep.x', 'C1.x')

        with self.assertRaises(NameError) as cm:
            prob.setup()

        self.assertEqual(str(cm.exception),
                         "'C1' <class ExecComp>: cannot use variable name 'units' because it's a reserved keyword.")

    def test_common_units(self):
        # all variables in the ExecComp have the same units
        prob = om.Problem()

        prob.model.add_subsystem('indep', om.IndepVarComp('x', 100.0, units='cm'))
        prob.model.add_subsystem('comp', om.ExecComp('y=x+z+1.', units='m',
                                                     x={'val': 2.0},
                                                     z=2.0))
        prob.model.connect('indep.x', 'comp.x')

        prob.setup()
        prob.run_model()

        assert_near_equal(prob['comp.y'], 4.0, 0.00001)

    def test_common_units_no_meta(self):
        # make sure common units are assigned when no metadata is provided
        prob = om.Problem()

        prob.model.add_subsystem('indep', om.IndepVarComp('x', 2.0, units='km'))
        prob.model.add_subsystem('comp', om.ExecComp('y = x+1', units='m'))

        prob.model.connect('indep.x', 'comp.x')

        prob.setup()
        prob.run_model()

        assert_near_equal(prob['comp.y'], 2001., 0.00001)

    def test_conflicting_units(self):
        prob = om.Problem()
        prob.model.add_subsystem('indep', om.IndepVarComp('x', 100.0, units='cm'))
        C1 = prob.model.add_subsystem('C1', om.ExecComp('y=x+z+1.', units='m',
                                                        x={'val': 2.0, 'units': 'km'},
                                                        z=2.0))
        prob.model.connect('indep.x', 'C1.x')

        with self.assertRaises(RuntimeError) as cm:
            prob.setup()

        self.assertEqual(str(cm.exception),
                         "'C1' <class ExecComp>: units of 'km' have been specified for variable 'x', but "
                         "units of 'm' have been specified for the entire component.")

    def test_shape_and_value(self):
        p = om.Problem()
        model = p.model

        model.add_subsystem('comp', om.ExecComp('y=3.0*x + 2.5',
                                                x={'shape': (5,), 'val': np.zeros(5)},
                                                y={'shape': (5,), 'val': np.zeros(5)}))

        p.setup()
        p.run_model()

        J = p.compute_totals(of=['comp.y'], wrt=['comp.x'], return_format='array')

        assert_almost_equal(J, np.eye(5)*3., decimal=6)

    def test_conflicting_shape(self):
        p = om.Problem()
        model = p.model

        model.add_subsystem('comp', om.ExecComp('y=3.0*x + 2.5',
                                                x={'shape': (5,), 'val': 5},
                                                y={'shape': (5,)}))

        with self.assertRaises(Exception) as context:
            p.setup()

        self.assertEqual(str(context.exception).replace('L,', ','),  # L on Windows
                         "'comp' <class ExecComp>: shape of (5,) has been specified for variable 'x', "
                         "but a value of shape (1,) has been provided.")

    def test_common_shape(self):
        p = om.Problem()
        model = p.model

        model.add_subsystem('comp', om.ExecComp('y=3.0*x + 2.5', shape=(5,)))

        p.setup()
        p.run_model()

        J = p.compute_totals(of=['comp.y'], wrt=['comp.x'], return_format='array')

        assert_almost_equal(J, np.eye(5)*3., decimal=6)

    def test_common_shape_with_values(self):
        p = om.Problem()
        model = p.model

        model.add_subsystem('comp', om.ExecComp('y=3.0*x + 2.5', shape=(5,),
                                                x={'val': np.zeros(5)},
                                                y={'val': np.zeros(5)}))

        p.setup()
        p.run_model()

        J = p.compute_totals(of=['comp.y'], wrt=['comp.x'], return_format='array')

        assert_almost_equal(J, np.eye(5)*3., decimal=6)

    def test_common_shape_conflicting_shape(self):
        p = om.Problem()
        model = p.model

        model.add_subsystem('comp', om.ExecComp('y=3.0*x + 2.5', shape=(5,),
                                                y={'shape': (10,)}))

        with self.assertRaises(Exception) as context:
            p.setup()

        self.assertEqual(str(context.exception).replace('L,', ','),  # L on Windows
                         "'comp' <class ExecComp>: shape of (10,) has been specified for variable 'y', "
                         "but shape of (5,) has been specified for the entire component.")

    def test_common_shape_conflicting_value(self):
        p = om.Problem()
        model = p.model

        model.add_subsystem('comp', om.ExecComp('y=3.0*x + 2.5', shape=(5,),
                                                x={'val': 5}))

        with self.assertRaises(Exception) as context:
            p.setup()

        self.assertEqual(str(context.exception).replace('1L,', '1,'),  # 1L on Windows
                         "'comp' <class ExecComp>: value of shape (1,) has been specified for variable 'x', "
                         "but shape of (5,) has been specified for the entire component.")

    def test_math(self):
        prob = om.Problem()
        C1 = prob.model.add_subsystem('C1', om.ExecComp('y=sin(x)', x=2.0))

        prob.setup()

        # Conclude setup but don't run model.
        prob.final_setup()

        self.assertTrue('x' in C1._inputs)
        self.assertTrue('y' in C1._outputs)

        prob.set_solver_print(level=0)
        prob.run_model()

        assert_near_equal(C1._outputs['y'], math.sin(2.0), 0.00001)

    def test_np(self):
        prob = om.Problem()
        prob.model.add_subsystem('C1', om.ExecComp(['y1=sin(x)', 'y2=np.cos(x)'], x=2.0))
        prob.setup()

        with self.assertRaises(Exception) as context:
          prob.run_model()

        self.assertEqual(str(context.exception),
            "'C1' <class ExecComp>: Error occurred evaluating 'y2=np.cos(x)':\n"
            "    ExecComp supports a subset of numpy functions directly, without the 'np' prefix.\n"
            "    'cos' is supported (See the documentation).")

    def test_numpy(self):
        prob = om.Problem()
        prob.model.add_subsystem('C1', om.ExecComp('y=numpy.sin(x)', x=2.0))
        prob.setup()

        with self.assertRaises(Exception) as context:
          prob.run_model()

        self.assertEqual(str(context.exception),
            "'C1' <class ExecComp>: Error occurred evaluating 'y=numpy.sin(x)':\n"
            "    ExecComp supports a subset of numpy functions directly, without the 'numpy' prefix.\n"
            "    'sin' is supported (See the documentation).")

    def test_numpy_fft(self):
        prob = om.Problem()
        prob.model.add_subsystem('C1', om.ExecComp('y=numpy.fft(x)', x=2.0))
        prob.setup()

        with self.assertRaises(Exception) as context:
          prob.run_model()

        self.assertEqual(str(context.exception),
            "'C1' <class ExecComp>: Error occurred evaluating 'y=numpy.fft(x)':\n"
            "    ExecComp supports a subset of numpy functions directly, without the 'numpy' prefix.\n"
            "    'fft' is not supported (See the documentation).")

    def test_array(self):
        prob = om.Problem()
        C1 = prob.model.add_subsystem('C1', om.ExecComp('y=x[1]',
                                                        x=np.array([1., 2., 3.]),
                                                        y=0.0))

        prob.setup()

        # Conclude setup but don't run model.
        prob.final_setup()

        self.assertTrue('x' in C1._inputs)
        self.assertTrue('y' in C1._outputs)

        prob.set_solver_print(level=0)
        prob.run_model()

        assert_near_equal(C1._outputs['y'], 2.0, 0.00001)

    def test_array_lhs(self):
        prob = om.Problem()
        C1 = prob.model.add_subsystem('C1', om.ExecComp(['y[0]=x[1]', 'y[1]=x[0]'],
                                                        x=np.array([1., 2., 3.]),
                                                        y=np.array([0., 0.])))

        prob.setup()

        # Conclude setup but don't run model.
        prob.final_setup()

        self.assertTrue('x' in C1._inputs)
        self.assertTrue('y' in C1._outputs)

        prob.set_solver_print(level=0)
        prob.run_model()

        assert_near_equal(C1._outputs['y'], np.array([2., 1.]), 0.00001)

    def test_simple_array_model(self):
        prob = om.Problem()
        prob.model.add_subsystem('comp', om.ExecComp(['y[0]=2.0*x[0]+7.0*x[1]',
                                                      'y[1]=5.0*x[0]-3.0*x[1]'],
                                                     x=np.zeros([2]), y=np.zeros([2])))

        prob.setup()
        prob.set_solver_print(level=0)
        prob.run_model()

        data = force_check_partials(prob, out_stream=None)

        assert_check_partials(data, atol=1e-5, rtol=1e-5)

    def test_simple_array_model2(self):
        prob = om.Problem()
        prob.model.add_subsystem('comp', om.ExecComp('y = mat.dot(x)',
                                                     x=np.zeros((2,)), y=np.zeros((2,)),
                                                     mat=np.array([[2., 7.], [5., -3.]])))

        prob.setup()
        prob.set_solver_print(level=0)
        prob.run_model()

        data = force_check_partials(prob, out_stream=None)

        assert_check_partials(data, atol=1e-5, rtol=1e-5)

    def test_complex_step(self):
        prob = om.Problem()
        C1 = prob.model.add_subsystem('C1', om.ExecComp(['y=2.0*x+1.'], x=2.0))

        prob.setup()

        # Conclude setup but don't run model.
        prob.final_setup()

        self.assertTrue('x' in C1._inputs)
        self.assertTrue('y' in C1._outputs)

        prob.set_solver_print(level=0)
        prob.run_model()

        assert_near_equal(C1._outputs['y'], 5.0, 0.00001)

        C1._linearize()

        assert_near_equal(C1._jacobian[('y', 'x')], [[2.0]], 0.00001)

    def test_complex_step2(self):
        prob = om.Problem(om.Group())
        prob.model.add_subsystem('comp', om.ExecComp('y=x*x + x*2.0', x=2.0))
        prob.set_solver_print(level=0)

        prob.setup(check=False, mode='fwd')
        prob.run_model()

        J = prob.compute_totals(['comp.y'], ['comp.x'], return_format='flat_dict')
        assert_near_equal(J['comp.y', 'comp.x'], np.array([[6.0]]), 0.00001)

        prob.setup(check=False, mode='rev')
        prob.run_model()

        J = prob.compute_totals(['comp.y'], ['comp.x'], return_format='flat_dict')
        assert_near_equal(J['comp.y', 'comp.x'], np.array([[6.0]]), 0.00001)

    def test_abs_complex_step(self):
        prob = om.Problem()
        C1 = prob.model.add_subsystem('C1', om.ExecComp('y=2.0*abs(x)', x=-2.0))

        prob.setup()
        prob.set_solver_print(level=0)
        prob.run_model()

        assert_near_equal(C1._outputs['y'], 4.0, 0.00001)

        # any positive C1.x should give a 2.0 derivative for dy/dx
        C1._inputs['x'] = 1.0e-10
        C1._linearize()
        assert_near_equal(C1._jacobian['y', 'x'], [[2.0]], 0.00001)

        C1._inputs['x'] = -3.0
        C1._linearize()
        assert_near_equal(C1._jacobian['y', 'x'], [[-2.0]], 0.00001)

        C1._inputs['x'] = 0.0
        C1._linearize()
        assert_near_equal(C1._jacobian['y', 'x'], [[2.0]], 0.00001)

    def test_arctan_complex_step(self):
        prob = om.Problem()
        C1 = prob.model.add_subsystem('C1', om.ExecComp('z=2.0*arctan2(y, x)', x=np.array([1+2j]), y=1))

        prob.setup()
        prob.set_solver_print(level=0)
        prob.run_model()

        assert_near_equal(C1._outputs['z'], np.array([1.57079633]), 1e-8)

    def test_abs_array_complex_step(self):
        prob = om.Problem()
        C1 = prob.model.add_subsystem('C1', om.ExecComp('y=2.0*abs(x)',
                                                        x=np.ones(3)*-2.0, y=np.zeros(3)))

        prob.setup()
        prob.set_solver_print(level=0)
        prob.run_model()

        assert_near_equal(C1._outputs['y'], np.ones(3)*4.0, 0.00001)

        # any positive C1.x should give a 2.0 derivative for dy/dx
        C1._inputs['x'] = np.ones(3)*1.0e-10
        C1._linearize()
        assert_near_equal(C1._jacobian['y', 'x'], np.eye(3)*2.0, 0.00001)

        C1._inputs['x'] = np.ones(3)*-3.0
        C1._linearize()
        assert_near_equal(C1._jacobian['y', 'x'], np.eye(3)*-2.0, 0.00001)

        C1._inputs['x'] = np.zeros(3)
        C1._linearize()
        assert_near_equal(C1._jacobian['y', 'x'], np.eye(3)*2.0, 0.00001)

        C1._inputs['x'] = np.array([1.5, -0.6, 2.4])
        C1._linearize()
        expect = np.zeros((3, 3))
        expect[0, 0] = 2.0
        expect[1, 1] = -2.0
        expect[2, 2] = 2.0

        assert_near_equal(C1._jacobian['y', 'x'], expect, 0.00001)

    def test_has_diag_partials_error(self):
        p = om.Problem()
        model = p.model

        mat = np.arange(15).reshape((3,5))
        model.add_subsystem('comp', om.ExecComp('y=A.dot(x)', has_diag_partials=True, A=mat,
                                                x=np.ones(5), y=np.ones(3)))

        p.setup()

        with self.assertRaises(Exception) as context:
            p.final_setup()

        self.assertEqual(str(context.exception),
                         "'comp' <class ExecComp>: has_diag_partials is True but partial(y, A) is not square (shape=(3, 15)).")

    def test_has_diag_partials(self):
        # Really check to see that the has_diag_partials argument had its intended effect

        # run with has_diag_partials=False
        p = om.Problem()
        model = p.model
        comp = om.ExecComp('y=3.0*x + 2.5', has_diag_partials=False, x=np.ones(5), y=np.ones(5))
        model.add_subsystem('comp', comp)
        p.setup()

        self.assertEqual(len(comp._declared_partials_patterns), 0)

        # run with has_diag_partials=True
        p = om.Problem()
        model = p.model
        comp = om.ExecComp('y=3.0*x + 2.5', has_diag_partials=True, x=np.ones(5), y=np.ones(5))
        model.add_subsystem('comp', comp)
        p.setup()
        p.final_setup()

        declared_partials = comp._declared_partials_patterns[('y','x')]
        self.assertTrue('rows' in declared_partials )
        self.assertListEqual([0,1,2,3,4], list( comp._declared_partials_patterns[('y','x')]['rows']))
        self.assertTrue('cols' in declared_partials )
        self.assertListEqual([0,1,2,3,4], list( comp._declared_partials_patterns[('y','x')]['cols']))

    def test_exec_comp_deriv_sparsity(self):
        # Check to make sure that when an ExecComp has more than one
        # expression that only the partials that are needed are declared and computed

        # with has_diag_partials set to the default of False and just scalars
        p = om.Problem()
        model = p.model
        comp = om.ExecComp(['y1=2.0*x1+1.', 'y2=3.0*x2-1.'],x1=1.0, x2=2.0)
        model.add_subsystem('comp', comp)
        p.setup()
        p.final_setup()

        ## make sure only the partials that are needed are declared
        declared_partials = comp._declared_partials_patterns
        self.assertListEqual( sorted([('y1', 'x1'), ('y2', 'x2') ]),
                              sorted(declared_partials.keys()))

        p.run_model()

        # make sure only what is needed was computed
        subjacs_info = comp._jacobian._subjacs_info
        self.assertListEqual(sorted([('comp.y1', 'comp.x1'), ('comp.y2', 'comp.x2'),
                                     ('comp.y1', 'comp.y1'),('comp.y2', 'comp.y2')]),
                             sorted(subjacs_info.keys()))

        # make sure the result of compute_partials is correct
        J = p.compute_totals(of=['comp.y1'], wrt=['comp.x1'], return_format='array')
        self.assertEqual(2.0, J)
        J = p.compute_totals(of=['comp.y2'], wrt=['comp.x2'], return_format='array')
        self.assertEqual(3.0, J)

        # make sure this works with arrays and when has_diag_partials is the default of False
        p = om.Problem()
        model = p.model
        comp = om.ExecComp(['y1=2.0*x1+1.', 'y2=3.0*x2-1.'],
                           x1=np.ones(5), y1=np.ones(5), x2=np.ones(5), y2=np.ones(5))
        model.add_subsystem('comp', comp)
        p.setup()
        p.final_setup()

        declared_partials = comp._declared_partials_patterns
        self.assertListEqual( sorted([('y1', 'x1'), ('y2', 'x2') ]),
                              sorted(declared_partials.keys()))

        p.run_model()
        J = p.compute_totals(of=['comp.y1'], wrt=['comp.x1'], return_format='array')
        self.assertTrue(np.all(2.0*np.identity(5) == J))
        J = p.compute_totals(of=['comp.y2'], wrt=['comp.x2'], return_format='array')
        self.assertTrue(np.all(3.0*np.identity(5) == J))

        # with has_diag_partials True to make sure that still works with arrays
        p = om.Problem()
        model = p.model
        comp = om.ExecComp(['y1=2.0*x1+1.', 'y2=3.0*x2-1.'], has_diag_partials=True,
                           x1=np.ones(5), y1=np.ones(5), x2=np.ones(5), y2=np.ones(5) )
        model.add_subsystem('comp', comp)
        p.setup()
        p.final_setup()

        declared_partials = comp._declared_partials_patterns
        self.assertListEqual( sorted([('y1', 'x1'), ('y2', 'x2') ]),
                              sorted(declared_partials.keys()))
        self.assertTrue('cols' in declared_partials[('y1', 'x1')] )
        self.assertTrue('rows' in declared_partials[('y1', 'x1')] )
        self.assertTrue('cols' in declared_partials[('y2', 'x2')] )
        self.assertTrue('rows' in declared_partials[('y2', 'x2')] )
        self.assertListEqual([0,1,2,3,4], list( comp._declared_partials_patterns[('y1','x1')]['rows']))
        self.assertListEqual([0,1,2,3,4], list( comp._declared_partials_patterns[('y1','x1')]['cols']))
        self.assertListEqual([0,1,2,3,4], list( comp._declared_partials_patterns[('y2','x2')]['rows']))
        self.assertListEqual([0,1,2,3,4], list( comp._declared_partials_patterns[('y2','x2')]['cols']))

        p.run_model()

        J = p.compute_totals(of=['comp.y1'], wrt=['comp.x1'], return_format='array')
        self.assertTrue(np.all(2.0*np.identity(5) == J))
        J = p.compute_totals(of=['comp.y2'], wrt=['comp.x2'], return_format='array')
        self.assertTrue(np.all(3.0*np.identity(5) == J))

    def test_has_diag_partials_shape_only(self):
        p = om.Problem()
        model = p.model

        model.add_subsystem('comp', om.ExecComp('y=3.0*x + 2.5', has_diag_partials=True,
                                                x={'shape': (5,)}, y={'shape': (5,)}))

        p.setup()
        p.run_model()

        J = p.compute_totals(of=['comp.y'], wrt=['comp.x'], return_format='array')

        assert_almost_equal(J, np.eye(5)*3., decimal=6)

    def test_tags(self):
        prob = om.Problem(model=om.Group())
        C1 = prob.model.add_subsystem('C1', om.ExecComp('y=x+z+1.',
                                                        x={'val': 1.0, 'units': 'm', 'tags': 'tagx'},
                                                        y={'units': 'm', 'tags': ['tagy','tagq']},
                                                        z={'val': 2.0, 'tags': 'tagz'}))

        prob.setup(check=False)

        prob.set_solver_print(level=0)
        prob.run_model()

        # Inputs no tags
        inputs = prob.model.list_inputs(val=False, prom_name=False, out_stream=None)
        self.assertEqual(sorted(inputs), [
            ('C1.x', {}),
            ('C1.z', {}),
        ])

        # Inputs with tags
        inputs = prob.model.list_inputs(val=False, prom_name=False, out_stream=None, tags="tagx")
        self.assertEqual(sorted(inputs), [
            ('C1.x', {}),
        ])

        # Inputs with multiple tags
        inputs = prob.model.list_inputs(val=False, prom_name=False, out_stream=None, tags=["tagx", "tagz"])
        self.assertEqual(sorted(inputs), [
            ('C1.x', {}),
            ('C1.z', {}),
        ])

        # Inputs with tag that does not match
        inputs = prob.model.list_inputs(val=False, prom_name=False, out_stream=None, tags="tag_wrong")
        self.assertEqual(sorted(inputs), [])

        # Outputs no tags
        outputs = prob.model.list_outputs(val=False, prom_name=False, out_stream=None)
        self.assertEqual(sorted(outputs), [
            ('C1.y', {}),
        ])

        # Outputs with tags
        outputs = prob.model.list_outputs(val=False, prom_name=False, out_stream=None, tags="tagy")
        self.assertEqual(sorted(outputs), [
            ('C1.y', {}),
        ])

        # Outputs with multiple tags
        outputs = prob.model.list_outputs(val=False, prom_name=False, out_stream=None, tags=["tagy", "tagx"])
        self.assertEqual(sorted(outputs), [
            ('C1.y', {}),
        ])

        # Outputs with tag that does not match
        outputs = prob.model.list_outputs(val=False, prom_name=False, out_stream=None, tags="tag_wrong")
        self.assertEqual(sorted(outputs), [])

    def test_feature_has_diag_partials(self):

        p = om.Problem()
        model = p.model

        model.add_subsystem('comp', om.ExecComp('y=3.0*x + 2.5',
                                                has_diag_partials=True,
                                                x=np.ones(5), y=np.ones(5)))

        p.setup()

        p.set_val('comp.x', np.ones(5))

        p.run_model()

        J = p.compute_totals(of=['comp.y'], wrt=['comp.x'], return_format='array')

        assert_almost_equal(J, np.eye(5)*3., decimal=6)

    def test_feature_multi_output(self):

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('comp', om.ExecComp(['y1=x+1.', 'y2=x-1.']), promotes=['x'])

        prob.setup()

        prob.set_val('x', 2.0)

        prob.set_solver_print(level=0)
        prob.run_model()

        assert_near_equal(prob.get_val('comp.y1'), 3.0, 0.00001)
        assert_near_equal(prob.get_val('comp.y2'), 1.0, 0.00001)

    def test_feature_multi_output2(self):
        # verify that expressions can have multiple LHS variables.

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('comp', om.ExecComp(['y1, y2 = x+1., x-1.']), promotes=['x'])

        prob.setup()

        prob.set_val('x', 2.0)

        prob.set_solver_print(level=0)
        prob.run_model()

        assert_near_equal(prob.get_val('comp.y1'), 3.0, 0.00001)
        assert_near_equal(prob.get_val('comp.y2'), 1.0, 0.00001)

    def test_feature_array(self):

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('comp', om.ExecComp('y=x[1]',
                                                x=np.array([1., 2., 3.]),
                                                y=0.0))

        prob.setup()

        prob.set_solver_print(level=0)
        prob.run_model()

        assert_near_equal(prob.get_val('comp.y'), 2.0, 0.00001)

    def test_feature_math(self):

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('comp', om.ExecComp('z = sin(x)**2 + cos(y)**2'))

        prob.setup()

        prob.set_val('comp.x', np.pi/2.0)
        prob.set_val('comp.y', np.pi/2.0)

        prob.set_solver_print(level=0)
        prob.run_model()

        assert_near_equal(prob.get_val('comp.z'), 1.0, 0.00001)

    def test_feature_numpy(self):

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('comp', om.ExecComp('y=sum(x)', x=np.array([1., 2., 3.])))

        prob.setup()

        prob.set_solver_print(level=0)
        prob.run_model()

        assert_near_equal(prob['comp.y'], 6.0, 0.00001)

    def test_feature_metadata(self):

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('comp', om.ExecComp('z=x+y',
                                                x={'val': 0.0, 'units': 'inch'},
                                                y={'val': 0.0, 'units': 'inch'},
                                                z={'val': 0.0, 'units': 'inch'}))

        prob.setup()

        prob.set_val('comp.x', 12.0, units='inch')
        prob.set_val('comp.y', 1.0, units='ft')

        prob.set_solver_print(level=0)
        prob.run_model()

        assert_near_equal(prob.get_val('comp.z'), 24.0, 0.00001)

    def test_feature_options(self):

        model = om.Group()

        xcomp = model.add_subsystem('comp', om.ExecComp('y=2*x', shape=(2,)))

        xcomp.options['units'] = 'm'

        prob = om.Problem(model)
        prob.setup()

        prob.set_val('comp.x', [100., 200.], units='cm')

        prob.run_model()

        assert_near_equal(prob.get_val('comp.y'), [2., 4.], 0.00001)

    def test_list_outputs_resids_tol(self):
        prob = om.Problem()
        model = prob.model

        model.add_subsystem(
            "quad_1",
            om.ExecComp(
                "y = a * x ** 2 + b * x + c",
                a={"val": 2.0},
                b={"val": 5.0},
                c={"val": 3.0},
                x={"shape": (2,)},
                y={"shape": (2,)},
            ),
        )

        balance = model.add_subsystem("balance", om.BalanceComp())
        balance.add_balance("x_1", val=np.array([1, -1]), rhs_val=np.array([0., 0.]))
        model.connect("balance.x_1", "quad_1.x")
        model.connect("quad_1.y", "balance.lhs:x_1")

        prob.model.linear_solver = om.ScipyKrylov()
        prob.model.nonlinear_solver = om.NewtonSolver(solve_subsystems=False, maxiter=100, iprint=2)

        prob.setup()
        prob.model.nonlinear_solver.options["maxiter"] = 0
        prob.run_model()

        stream = StringIO()
        outputs = prob.model.list_outputs(residuals=True, residuals_tol=1e-5, out_stream=stream)

        text = stream.getvalue()
        self.assertTrue("balance" in text)
        self.assertTrue("x_1" in text)

    def test_add_expr(self):
        p = om.Problem()

        excomp = om.ExecComp('y=x',
                             x={'val' : 3.0, 'units' : 'mm'},
                             y={'shape' : (1, ), 'units' : 'cm'})

        excomp.add_expr('z = 2.9*x',
                        z={'shape' : (1, ), 'units' : 's'})

        p.model.add_subsystem('comp', excomp, promotes=['*'])
        p.setup()
        p.run_model()

        assert_almost_equal(p.get_val('z'), 8.7, 1e-8)
        assert_almost_equal(p.get_val('y'), 3.0, 1e-8)

    def test_add_expr_bare(self):
        p = om.Problem()

        excomp = om.ExecComp()

        excomp.add_expr('z = 2.9*x',
                        x={'val' : 3.0, 'units' : 'mm'},
                        z={'shape' : (1, ), 'units' : 's'})

        p.model.add_subsystem('comp', excomp, promotes=['*'])
        p.setup()
        p.run_model()

        assert_almost_equal(p.get_val('z'), 8.7, 1e-8)

    def test_add_expr_configure(self):

        class ConfigGroup(om.Group):
            def setup(self):
                excomp = om.ExecComp('y=x',
                                     x={'val' : 3.0, 'units' : 'mm'},
                                     y={'shape' : (1, ), 'units' : 'cm'})

                self.add_subsystem('excomp', excomp, promotes=['*'])

            def configure(self):
                self.excomp.add_expr('z = 2.9*x',
                                     z={'shape' : (1, ), 'units' : 's'})


        p = om.Problem()
        p.model.add_subsystem('sub', ConfigGroup(), promotes=['*'])
        p.setup()
        p.run_model()

        assert_almost_equal(p.get_val('z'), 8.7, 1e-8)
        assert_almost_equal(p.get_val('y'), 3.0, 1e-8)

    def test_add_expr_configure_no_shape_or_val(self):

        class ConfigGroup(om.Group):
            def setup(self):
                excomp = om.ExecComp('x_constraint=x',
                                     x_constraint={'units': None},
                                     x={'units': 's'})

                self.add_subsystem('excomp', excomp, promotes=['*'])

            def configure(self):
                self.excomp.add_expr('y_constraint = y',
                                     y_constraint={'units': None},
                                     y={'units': 's'})

        p = om.Problem()
        p.model.add_subsystem('sub', ConfigGroup(), promotes=['*'])
        p.setup()
        p.set_val('x', 3.0)
        p.set_val('y', 4.0)
        p.run_model()

        assert_almost_equal(p.get_val('x_constraint'), 3.0)
        assert_almost_equal(p.get_val('y_constraint'), 4.0)

    def test_no_expressions_until_configure(self):

        class ConfigGroup(om.Group):
            def setup(self):
                excomp = om.ExecComp()

                self.add_subsystem('excomp', excomp, promotes=['*'])

            def configure(self):
                self.excomp.add_expr('y_constraint = y**2',
                                     y_constraint={'units': None},
                                     y={'units': 's'})

        p = om.Problem()
        p.model.add_subsystem('sub', ConfigGroup(), promotes=['*'])
        p.setup()
        p.set_val('y', 4.0)
        p.run_model()

        assert_almost_equal(p.get_val('y_constraint'), 16.0)

    def test_no_expressions(self):

        class ConfigGroup(om.Group):
            def setup(self):
                excomp = om.ExecComp()

                self.add_subsystem('excomp', excomp, promotes=['*'])

            def configure(self):
                pass

        p = om.Problem()
        p.model.add_subsystem('sub', ConfigGroup(), promotes=['*'])
        p.setup()
        p.run_model()

    def test_add_expr_configure_delay_defaults(self):

        class ConfigGroup(om.Group):
            def setup(self):
                excomp = om.ExecComp('y=x',
                                     y={'shape' : (1, ), 'units' : 'cm'})

                self.add_subsystem('excomp', excomp, promotes=['*'])

            def configure(self):
                self.excomp.add_expr('z = 2.9*x',
                                     x={'val' : 3.0, 'units' : 'mm'},
                                     z={'shape' : (1, ), 'units' : 's'})


        p = om.Problem()
        p.model.add_subsystem('sub', ConfigGroup(), promotes=['*'])
        p.setup()
        p.run_model()

        assert_almost_equal(p.get_val('z'), 8.7, 1e-8)
        assert_almost_equal(p.get_val('y'), 3.0, 1e-8)

    def test_add_expr_errors(self):
        p = om.Problem()

        excomp = om.ExecComp('y=x',
                             x={'val' : 3.0, 'units' : 'mm'},
                             y={'shape' : (1, ), 'units' : 'cm'})

        with self.assertRaises(NameError) as cm:
            excomp.add_expr('z = 2.9*x',
                            x={'val' : 3.0, 'units' : 'cm'},
                            z={'shape' : (1, ), 'units' : 's'})

        self.assertEqual(cm.exception.args[0],
                         "Defaults for 'x' have already been defined in a previous "
                         "expression.")

        with self.assertRaises(TypeError) as cm:
            excomp.add_expr(p)

        self.assertEqual(cm.exception.args[0],
                         "Argument 'expr' must be of type 'str', but type 'Problem' was found.")

        excomp.add_expr('y = 2.9*x')
        p.model.add_subsystem('zzz', excomp)
        with self.assertRaises(RuntimeError) as cm:
            p.setup()

        self.assertEqual(cm.exception.args[0],
                         "'zzz' <class ExecComp>: The output 'y' has already been defined by an expression.")

    def test_feature_add_expr(self):

        class ConfigGroup(om.Group):
            def setup(self):
                excomp = om.ExecComp('y=x',
                                     x={'val' : 3.0, 'units' : 'mm'},
                                     y={'shape' : (1, ), 'units' : 'cm'})

                self.add_subsystem('excomp', excomp, promotes=['*'])

            def configure(self):
                self.excomp.add_expr('z = 2.9*x',
                                     z={'shape' : (1, ), 'units' : 's'})

        p = om.Problem()
        p.model.add_subsystem('sub', ConfigGroup(), promotes=['*'])
        p.setup()
        p.run_model()

        assert_almost_equal(p.get_val('z'), 8.7, 1e-8)
        assert_almost_equal(p.get_val('y'), 3.0, 1e-8)

    def test_val_type(self):
        class ConfigGroup(om.Group):
            def setup(self):
                excomp = om.ExecComp('y=x',
                                     x={'val': 3.0, 'units': 'mm'},
                                     y={'shape': (1,), 'units': 'cm'})

                self.add_subsystem('excomp', excomp, promotes=['*'])

            def configure(self):
                self.excomp.add_expr('z = 2.9*x',
                                     z={'shape': (1,), 'units': 's'})

        p = om.Problem()
        p.model.add_subsystem('sub', ConfigGroup(), promotes=['*'])
        p.setup()

        # check that 'val' has been saved as a numpy array (see issue #2415)
        abs2meta = p.model._var_abs2meta
        for typ in ['input', 'output']:
            for abs_name in abs2meta[typ]:
                val = abs2meta[typ][abs_name]['val']
                if not isinstance(val, np.ndarray):
                    self.fail(f"The {typ} variable '{abs_name}' has a value of {val} "
                              f"with type {type(val)}, but it should be a numpy array.")


    def test_constants_simple(self):
        prob = om.Problem()
        C1 = prob.model.add_subsystem('C1', om.ExecComp('x = a + b', a={'val': 6, 'constant':True}))

        prob.setup()
        prob.final_setup()

        self.assertFalse('a' in C1._inputs)
        self.assertTrue('b' in C1._inputs)
        self.assertTrue('x' in C1._outputs)

        prob.set_solver_print(level=0)
        prob.run_model()

        assert_near_equal(C1._outputs['x'], 7.0, 0.00001)

    def test_constants_val_not_given(self):
        prob = om.Problem()
        prob.model.add_subsystem('C1', om.ExecComp('x = a + b', a={'constant':True}))

        with self.assertRaises(RuntimeError) as cm:
            prob.setup()
            self.assertEqual(str(cm.exception), "C1' <class ExecComp>: arg 'a' in call to ExecComp() is a constant but no value is given")

    def test_constants_units_given(self):
        prob = om.Problem()
        prob.model.add_subsystem('C1', om.ExecComp('x = a + b', a={'constant':True, 'val': 6, 'units': 'ft'}))

        expected_warning_msg = "'C1' <class ExecComp>: arg 'a' in call to " \
                               "ExecComp() is a constant. The units will be ignored"
        with assert_warning(SetupWarning, expected_warning_msg):
            prob.setup()

    def test_constants_shape_given(self):
        prob = om.Problem()
        prob.model.add_subsystem('C1', om.ExecComp('x = a + b', a={'constant':True, 'val': 6, 'shape': (1,)}))

        expected_warning_msg = "'C1' <class ExecComp>: arg 'a' in call to " \
                               "ExecComp() is a constant. The shape will be ignored"
        with assert_warning(SetupWarning, expected_warning_msg):
            prob.setup()

    def test_constants_set_constant_false(self):
        prob = om.Problem()
        C1 = prob.model.add_subsystem('C1', om.ExecComp('x = a + b', a={'val': 6, 'constant':False}))

        prob.setup()

        # Conclude setup but don't run model.
        prob.final_setup()

        self.assertTrue('a' in C1._inputs)
        self.assertTrue('b' in C1._inputs)
        self.assertTrue('x' in C1._outputs)

        prob.set_solver_print(level=0)
        prob.run_model()

        assert_near_equal(C1._outputs['x'], 7.0, 0.00001)

    def test_constants_multiple_equations(self):
        prob = om.Problem()
        model = prob.model
        comp = om.ExecComp(['y1=2.0*x1+a', 'y2=3.0*x2-a'], x1=1.0, x2=2.0, a={'val': 6, 'constant':True})
        model.add_subsystem('comp', comp)
        prob.setup()
        prob.run_model()

        assert_near_equal(comp._outputs['y1'], 8.0, 0.00001)
        assert_near_equal(comp._outputs['y2'], 0.0, 0.00001)

    def test_constants_arrays(self):
        prob = om.Problem()
        model = prob.model

        comp = model.add_subsystem('comp', om.ExecComp('y=x+a',
                                                       y={'shape': (3,), 'val': np.zeros(3)},
                                                       x=np.array([1., 2., 3.]),
                                                       a={ 'val': np.array([1., 2., 3.]), 'constant':True}))

        prob.setup()

        prob.set_solver_print(level=0)
        prob.run_model()

        assert_near_equal(comp._outputs['y'], 2.0 * np.array([1., 2., 3.]), 0.00001)


class TestFunctionRegistration(unittest.TestCase):

    # These 2 tests don't run normally unless you run testflo with a -m "featuretest_*"
    # I don't want to include them in our normal tests because they don't contain
    # the _temporary_expr_dict contextmanager (to avoid user confusion)
    # which keeps the contents of the ExecComp _expr_dict clean
    def featuretest_register_simple(self):

        om.ExecComp.register("myfunc", lambda x: x * x, complex_safe=True)
        p = om.Problem()
        comp = p.model.add_subsystem("comp", om.ExecComp("y = 2 * myfunc(x)"))

        p.setup()
        p.run_model()
        J = p.compute_totals(of=['comp.y'], wrt=['comp.x'])
        assert_near_equal(J['comp.y', 'comp.x'][0][0], 4., 1e-10)

    def featuretest_register_simple_unsafe(self):

        # the following function isn't really complex unsafe, but we'll call it unsafe anyway
        # for demonstration purposes
        om.ExecComp.register("unsafe", lambda x: x * x, complex_safe=False)
        p = om.Problem()
        comp = p.model.add_subsystem("comp", om.ExecComp("y = 2 * unsafe(x)"))

        # because our function is complex unsafe, we must declare that the partials
        # with respect to 'x' use 'fd' instead of 'cs'
        comp.declare_partials('*', 'x', method='fd')

        p.setup()
        p.run_model()
        J = p.compute_totals(of=['comp.y'], wrt=['comp.x'])
        assert_near_equal(J['comp.y', 'comp.x'][0][0], 4., 1e-5)

    def test_register_simple(self):
        with _temporary_expr_dict():
            om.ExecComp.register('area', lambda x: x**2, complex_safe=True)
            p = om.Problem()
            p.model.add_subsystem('comp', om.ExecComp('area_square = area(x)'))
            p.setup()
            p['comp.x'] = 3.
            p.run_model()
            self.assertEqual(p['comp.area_square'], 9.)
            J = p.compute_totals(of=['comp.area_square'], wrt=['comp.x'])
            self.assertEqual(J['comp.area_square', 'comp.x'], 6.)

    def test_register_simple_arr(self):
        with _temporary_expr_dict():
            size = 5
            om.ExecComp.register('area', lambda x: x**2, complex_safe=True)
            p = om.Problem()
            p.model.add_subsystem('comp', om.ExecComp('area_square = area(x)', shape=size))
            p.setup()
            p['comp.x'] = x = np.arange(1, size+1, dtype=float)
            p.run_model()
            assert_near_equal(p['comp.area_square'], x * x, 1e-11)
            J = p.compute_totals(of=['comp.area_square'], wrt=['comp.x'])
            assert_near_equal(J['comp.area_square', 'comp.x'], np.eye(size) * x * 2., 1e-11)

    def test_register_check_partials_not_safe(self):
        with _temporary_expr_dict():
            size = 10
            om.ExecComp.register('area', lambda x: x**2, complex_safe=False)
            p = om.Problem()
            comp = p.model.add_subsystem('comp', om.ExecComp('area_square = area(x)', shape=size))
            comp.declare_partials('*', '*', method='fd')
            p.setup()
            p['comp.x'] = 3.
            p.run_model()
            assert_near_equal(p['comp.area_square'], np.ones(size) * 9., 1e-6)

            data = force_check_partials(p, out_stream=None, step=1e-7)
            self.assertEqual(list(data), ['comp'])

    def test_register_check_partials_not_safe_err(self):
        with _temporary_expr_dict():
            size = 10
            om.ExecComp.register('area', lambda x: x**2, complex_safe=False)
            p = om.Problem()
            comp = p.model.add_subsystem('comp', om.ExecComp('area_square = area(x)', shape=size))
            p.setup()
            p['comp.x'] = 3.
            # calling run_model should NOT raise an exception
            p.run_model()
            assert_near_equal(p['comp.area_square'], np.ones(size) * 9., 1e-6)

            with self.assertRaises(Exception) as cm:
                data = force_check_partials(p, out_stream=None)
            self.assertEqual(cm.exception.args[0],
                             "'comp' <class ExecComp>: expression contains functions ['area'] that are not complex safe. To fix this, call declare_partials('*', ['x'], method='fd') on this component prior to setup.")

    def test_register_check_partials_not_safe_mult_expr(self):
        with _temporary_expr_dict():
            size = 5
            om.ExecComp.register('unsafe', lambda x: x**2, complex_safe=False)
            om.ExecComp.register('safe', lambda x: x**2, complex_safe=True)
            p = om.Problem()
            comp = p.model.add_subsystem('comp', om.ExecComp(['out1 = unsafe(x) * z',
                                                              'out2 = safe(y) + z'], shape=size))
            rows = cols = np.arange(size)
            comp.declare_partials('out1', ['x', 'z'], rows=rows, cols=cols, method='fd')
            comp.declare_partials('out2', ['y'], rows=rows, cols=cols, method='cs')
            comp.declare_partials('out2', ['z'], rows=rows, cols=cols, method='fd')
            p.setup()
            xx = np.arange(1, size + 1, dtype=float)
            p['comp.x'] = x = xx * 3.
            p['comp.y'] = y = xx * 4.
            p['comp.z'] = z = xx * 5.
            p.run_model()
            assert_near_equal(p['comp.out1'], x * x * z, 1e-10)
            assert_near_equal(p['comp.out2'], y * y + z, 1e-10)

            J = p.compute_totals(of=['comp.out1', 'comp.out2'], wrt=['comp.x', 'comp.y', 'comp.z'])
            assert_near_equal(J['comp.out1', 'comp.x'], np.eye(size) * 2. * x * z, 1e-6)
            assert_near_equal(J['comp.out1', 'comp.y'], np.zeros((size, size)), 1e-6)
            assert_near_equal(J['comp.out1', 'comp.z'], np.eye(size) * x**2, 1e-6)
            assert_near_equal(J['comp.out2', 'comp.x'], np.zeros((size, size)), 1e-11)
            assert_near_equal(J['comp.out2', 'comp.y'], np.eye(size) * 2. * y, 1e-11)
            assert_near_equal(J['comp.out2', 'comp.z'], np.eye(size), 1e-6)

            data = force_check_partials(p, out_stream=None, step=1e-7)
            self.assertEqual(list(data), ['comp'])

    def test_register_check_partials_not_safe_mult_expr_err(self):
        with _temporary_expr_dict():
            size = 10
            om.ExecComp.register('unsafe', lambda x: x**2, complex_safe=False)
            om.ExecComp.register('safe', lambda x: x**2, complex_safe=True)
            p = om.Problem()
            comp = p.model.add_subsystem('comp', om.ExecComp(['out1 = unsafe(x) * z',
                                                              'out2 = safe(y) + z'], shape=size))
            comp.declare_partials('*', ['x'], method='fd')
            p.setup()
            p['comp.x'] = 3.
            p['comp.y'] = 4.
            p['comp.z'] = 5.
            p.run_model()
            assert_near_equal(p['comp.out1'], np.ones(size) * 45., 1e-6)
            assert_near_equal(p['comp.out2'], np.ones(size) * 21., 1e-6)

            with self.assertRaises(Exception) as cm:
                data = force_check_partials(p, out_stream=None, step=1e-7)
            self.assertEqual(cm.exception.args[0],
                             "'comp' <class ExecComp>: expression contains functions ['unsafe'] that are not complex safe. To fix this, call declare_partials('*', ['z'], method='fd') on this component prior to setup.")

    def test_register_check_partials_safe(self):
        with _temporary_expr_dict():
            size = 10
            om.ExecComp.register('area', lambda x: x**2, complex_safe=True)
            p = om.Problem()
            p.model.add_subsystem('comp', om.ExecComp('area_square = area(x)', shape=size))
            p.setup()
            p['comp.x'] = 3.
            p.run_model()
            assert_near_equal(p['comp.area_square'], np.ones(size) * 9., 1e-11)

            data = force_check_partials(p, out_stream=None)
            self.assertEqual(list(data), ['comp'])

    def test_register_simple_arr_manual_partials_cs(self):
        with _temporary_expr_dict():
            size = 10
            om.ExecComp.register('area', lambda x: x**2, complex_safe=True)
            p = om.Problem()
            comp = p.model.add_subsystem('comp', om.ExecComp('area_square = area(x)', shape=size))
            comp.declare_partials('area_square', 'x', method='cs')
            p.setup()
            p['comp.x'] = 3.
            p.run_model()
            assert_near_equal(p['comp.area_square'], np.ones(size) * 9., 1e-11)
            J = p.compute_totals(of=['comp.area_square'], wrt=['comp.x'])
            assert_near_equal(J['comp.area_square', 'comp.x'], np.eye(size) * 6., 1e-11)

    def test_register_simple_arr_manual_partials_fd(self):
        with _temporary_expr_dict():
            size = 10
            om.ExecComp.register('area', lambda x: x**2, complex_safe=False)
            p = om.Problem()
            comp = p.model.add_subsystem('comp', om.ExecComp('area_square = area(x)', shape=size))
            comp.declare_partials('area_square', 'x', method='fd')
            p.setup()
            p['comp.x'] = 3.
            p.run_model()
            assert_near_equal(p['comp.area_square'], np.ones(size) * 9., 1e-6)
            J = p.compute_totals(of=['comp.area_square'], wrt=['comp.x'])
            assert_near_equal(J['comp.area_square', 'comp.x'], np.eye(size) * 6., 1e-6)

    def test_register_simple_arr_diag(self):
        with _temporary_expr_dict():
            size = 10
            om.ExecComp.register('area', lambda x: x**2, complex_safe=True)
            p = om.Problem()
            p.model.add_subsystem('comp', om.ExecComp('area_square = area(x)', shape=size, has_diag_partials=True))
            p.setup()
            p['comp.x'] = 3.
            p.run_model()
            assert_near_equal(p['comp.area_square'], np.ones(size) * 9., 1e-11)
            J = p.compute_totals(of=['comp.area_square'], wrt=['comp.x'])
            assert_near_equal(J['comp.area_square', 'comp.x'], np.eye(size) * 6., 1e-11)

            # verify diagonal subjac
            self.assertTrue(np.all(p.model.comp._subjacs_info['comp.area_square', 'comp.x']['rows'] == np.arange(size)))
            self.assertTrue(np.all(p.model.comp._subjacs_info['comp.area_square', 'comp.x']['cols'] == np.arange(size)))

    def test_register_shape_by_conn(self):
        with _temporary_expr_dict():
            size = 10
            om.ExecComp.register('part', lambda x: x[2:], complex_safe=True)
            p = om.Problem()
            p.model.add_subsystem('indeps', om.IndepVarComp('x', np.ones(size)))
            p.model.add_subsystem('comp', om.ExecComp('y = part(x) * 3.', shape_by_conn=True))
            p.model.add_subsystem('sink', om.ExecComp('y=x', shape=size-2))
            p.model.connect("indeps.x", "comp.x")
            p.model.connect("comp.y", "sink.x")
            p.setup()
            p['indeps.x'] = 3.
            p.run_model()
            assert_near_equal(p['comp.x'], np.ones(size) * 3., 1e-11)
            assert_near_equal(p['comp.y'], np.ones(size-2) * 9., 1e-11)
            assert_near_equal(p['sink.y'], np.ones(size-2) * 9., 1e-11)
            J = p.compute_totals(of=['sink.y'], wrt=['indeps.x'])
            assert_near_equal(J['sink.y', 'indeps.x'], np.eye(size)[2:, :] * 3., 1e-11)

    def test_register_shape_by_conn_err(self):
        with _temporary_expr_dict():
            size = 10
            om.ExecComp.register('double', lambda x: x * 2, complex_safe=True)
            p = om.Problem()
            p.model.add_subsystem('indeps', om.IndepVarComp('x', np.ones(size)))
            p.model.add_subsystem('comp', om.ExecComp('y = double(x) * 3.', shape_by_conn=True))
            p.model.add_subsystem('sink', om.ExecComp('y=x', shape=size-2))
            p.model.connect("indeps.x", "comp.x")
            p.model.connect("comp.y", "sink.x")
            p.setup()
            p['indeps.x'] = 3.

            # if shape_by_conn results in inputs and outputs being different shapes, we can't detect if that's
            # an error or not because we don't know how the expressions evaluate, so we have to wait until
            # compute runs and just report as an error during expression evaluation.

            # have to use regex to handle differences in numpy print formats for shape
            msg = "'comp' <class ExecComp>: Error occurred evaluating 'y = double\(x\) \* 3\.':\n" \
                  "could not broadcast input array from shape \(10.*\) into shape \(8.*\)"
            with self.assertRaisesRegex(Exception, msg) as cm:
                p.run_model()

    def test_shape_by_conn_bug_has_diag_partials_bug(self):
        # this is for a bug where has_diag_partials was being ignored when shape_by_conn
        # and/or copy_shape was used.

        prob = om.Problem()

        size = 100000
        t = np.linspace(0, 1, size)

        prob.model.add_subsystem("ivc",
                                om.IndepVarComp("t", val=t),
                                promotes_outputs=["*"])

        comp1 = prob.model.add_subsystem("comp1", om.ExecComp(["x = t + 1"],
                t={"shape_by_conn":True},
                x={"shape_by_conn":True, "copy_shape":"t"},
                has_diag_partials=True), promotes=["*"])

        comp2 = prob.model.add_subsystem("comp2", om.ExecComp(["y = t + 2"],
                t={"shape":size},
                y={"copy_shape":"t"},
                has_diag_partials=True), promotes=["*"])

        comp3 = prob.model.add_subsystem("comp3", om.ExecComp(["z = t + 3"],
                t={"shape_by_conn":True},
                z={"shape":size},
                has_diag_partials=True), promotes=["*"])

        comp4 = prob.model.add_subsystem("comp4", om.ExecComp(["w = t + 4"],
                t={"shape":size},
                w={"shape":size},
                has_diag_partials=True), promotes=["*"])

        prob.setup()
        prob.final_setup()

        # all subjac values should be size == size from above instead of (size, size)
        self.assertEqual(comp1._subjacs_info[('comp1.x', 'comp1.t')]['val'].size, size)
        self.assertEqual(comp2._subjacs_info[('comp2.y', 'comp2.t')]['val'].size, size)
        self.assertEqual(comp3._subjacs_info[('comp3.z', 'comp3.t')]['val'].size, size)
        self.assertEqual(comp4._subjacs_info[('comp4.w', 'comp4.t')]['val'].size, size)

    def test_register_err_keyword(self):
        with _temporary_expr_dict():
            with self.assertRaises(Exception) as cm:
                om.ExecComp.register('shape', lambda x: x, complex_safe=True)
            self.assertEqual(cm.exception.args[0], "ExecComp: cannot register name 'shape' because "
                             "it's a reserved keyword.")

    def test_register_err_not_callable(self):
        with _temporary_expr_dict():
            with self.assertRaises(Exception) as cm:
                om.ExecComp.register('foo', 99, complex_safe=True)
            self.assertEqual(cm.exception.args[0], "ExecComp: 'foo' passed to register() of type 'int' is not callable.")

    def test_register_err_dup(self):
        with _temporary_expr_dict():
            with self.assertRaises(Exception) as cm:
                om.ExecComp.register('exp', lambda x: x, complex_safe=True)
            self.assertEqual(cm.exception.args[0], "ExecComp: 'exp' has already been registered.")


_MASK = np.array(
    [[1, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 1],
    [0, 1, 0, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1],
    [0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1],
    [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0],
    [0, 1, 0, 1, 0, 1, 1, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0],
    [0, 1, 0, 1, 1, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1],
    [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1],
    [0, 1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 1, 0, 0, 1, 0, 0, 1, 0],
    [0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0],
    [1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0],
    [0, 1, 1, 0, 1, 1, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1]]
)


def setup_sparsity(mask):
    sparsity = np.random.random(mask.shape) + 1e-5
    return sparsity * mask



class TestFunctionRegistrationColoring(unittest.TestCase):
    def setUp(self):
        np.random.seed(11)
        self.startdir = os.getcwd()
        self.tempdir = tempfile.mkdtemp(prefix=self.__class__.__name__ + '_')
        os.chdir(self.tempdir)

    def tearDown(self):
        os.chdir(self.startdir)
        try:
            shutil.rmtree(self.tempdir)
        except OSError:
            pass

    def test_manual_coloring(self):
        with _temporary_expr_dict():

            prob = om.Problem(coloring_dir=self.tempdir)
            model = prob.model

            sparsity = setup_sparsity(_MASK)

            def mydot(x):
                return sparsity.dot(x)

            om.ExecComp.register('mydot', mydot, complex_safe=True)

            comp = model.add_subsystem('comp', om.ExecComp('y=mydot(x)',
                                                            x=np.ones(sparsity.shape[1]),
                                                            y=np.ones(sparsity.shape[0])))
            comp.declare_coloring('x', method='cs')

            prob.setup(mode='fwd')
            prob.set_solver_print(level=0)
            prob.run_model()

            J = prob.compute_totals('comp.y', 'comp.x')

            assert_near_equal(J['comp.y', 'comp.x'], sparsity)

            self.assertTrue(np.all(comp._coloring_info['coloring'].get_dense_sparsity() == _MASK))

    def test_auto_coloring(self):
        with _temporary_expr_dict():

            prob = om.Problem(coloring_dir=self.tempdir)
            model = prob.model

            sparsity = setup_sparsity(_MASK)

            def mydot(x):
                return sparsity.dot(x)

            om.ExecComp.register('mydot', mydot, complex_safe=True)

            comp = model.add_subsystem('comp', om.ExecComp('y=mydot(x)',
                                                            x=np.ones(sparsity.shape[1]),
                                                            y=np.ones(sparsity.shape[0])))
            prob.setup(mode='fwd')
            prob.set_solver_print(level=0)
            prob.run_model()

            J = prob.compute_totals('comp.y', 'comp.x')

            assert_near_equal(J['comp.y', 'comp.x'], sparsity)

            self.assertTrue(np.all(comp._coloring_info['coloring'].get_dense_sparsity() == _MASK))


class TestExecCompParameterized(unittest.TestCase):

    @parameterized.expand(itertools.product([
      func_name for func_name in _expr_dict if func_name not in ('np', 'numpy', '__builtins__')
    ]), name_func=lambda f, n, p: 'test_exec_comp_value_' + '_'.join(a for a in p.args))
    def test_exec_comp_value(self, f):
        test_data = _ufunc_test_data[f]

        prob = om.Problem()
        model = prob.model

        if len(test_data['args']) > 1:
            ivc = model.add_subsystem(name='ivc', subsys=om.IndepVarComp())
            for arg_name, arg_value in test_data['args'].items():
                if arg_name == 'f':
                    continue
                ivc.add_output(name=arg_name, val=arg_value['val'])
                model.connect('ivc.{0}'.format(arg_name), 'comp.{0}'.format(arg_name))

        model.add_subsystem('comp', om.ExecComp(test_data['str'], **test_data['args']),
                            promotes_outputs=['f'])
        prob.setup()

        if 'error' in test_data:
            err, msg = test_data['error']
            with self.assertRaises(err) as cm:
                prob.run_model()
            self.assertTrue(msg in str(cm.exception))
            return
        elif 'warning' in test_data:
            with assert_warning(*test_data['warning']):
                prob.run_model()
        else:
            prob.run_model()

        if 'check_func' in test_data:
            check_args = []
            try:
                check_args.append(test_data['args']['x']['val'])
            except Exception:
                pass
            try:
                check_args.append(test_data['args']['y']['val'])
            except Exception:
                pass
            check_args = tuple(check_args)

            expected = test_data['check_func'](*check_args)
        else:
            expected = test_data['check_val']
        np.testing.assert_almost_equal(prob['f'], expected)

        if 'check_val' not in test_data:
            try:
                force_check_partials(prob, out_stream=None)
            except TypeError as e:
                print(f, 'does not support complex-step differentiation')

    @parameterized.expand(itertools.product([
      func_name for func_name in _expr_dict if func_name not in ('np', 'numpy', '__builtins__')
    ]), name_func=lambda f, n, p: 'test_exec_comp_jac_' + '_'.join(a for a in p.args))
    def test_exec_comp_jac(self, f):
        test_data = _ufunc_test_data[f]

        prob = om.Problem()
        model = prob.model

        if len(test_data['args']) > 1:
            ivc = model.add_subsystem(name='ivc', subsys=om.IndepVarComp())
            for arg_name, arg_value in test_data['args'].items():
                if arg_name == 'f':
                    continue
                ivc.add_output(name=arg_name, val=arg_value['val'])
                model.connect('ivc.{0}'.format(arg_name),
                              '{0}_comp.{1}'.format(f, arg_name))

        model.add_subsystem('{0}_comp'.format(f),
                            om.ExecComp(test_data['str'], **test_data['args']),
                            promotes_outputs=['f'])
        prob.setup()

        if 'error' in test_data:
            err, msg = test_data['error']
            with self.assertRaises(err) as cm:
                prob.run_model()
            self.assertTrue(msg in str(cm.exception))
            return
        elif 'warning' in test_data:
            with assert_warning(*test_data['warning']):
                prob.run_model()
        else:
            prob.run_model()

        if 'check_val' not in test_data:
            cpd = force_check_partials(prob, out_stream=None)

            for comp in cpd:
                for (var, wrt) in cpd[comp]:
                    np.testing.assert_almost_equal(cpd[comp][var, wrt]['abs error'][0], 0, decimal=4)


if __name__ == "__main__":
    unittest.main()



"""
Functions for handling runtime function hooks.
"""

import sys
import warnings
import weakref
from inspect import getmro

from openmdao.utils.om_warnings import issue_warning


# global dict of hooks
# {class_name: { inst_id: {fname: [pre_hooks, post_hooks]}}}
# At class_name level, there might be a 'None' entry, which means that all instances of
# that class will have have the hooks specified in the 'None' dict.  Otherwise, entries
# there will be specific instance names.
# pre_hooks and post_hooks are lists of functions to call before or after the function named
# 'fname'.
_hooks = {}

# use this to keep hook ordering consistent across multiple calls to register_hook
_hook_counter = 0

# global switch that turns hook machinery on/off. Need it on in general for the default
# reporting system
use_hooks = True


def _reset_all_hooks():
    global _hooks
    _hooks = {}


class _HookMeta(object):
    def __init__(self, class_name, inst_id, hook, ncalls=None, exit=False, **kwargs):
        global _hook_counter
        self._stamp = _hook_counter
        _hook_counter += 1

        self.class_name = class_name
        self.inst_id = inst_id
        self.hook = hook
        self.ncalls = ncalls
        self.exit = exit
        self.kwargs = kwargs
        self.children = []  # if we're a 'None' inst_id hook, keep track of our child hooks

    def __repr__(self):
        return f"<_HookMeta {self.class_name} {self.inst_id} {self.hook} {self.ncalls} "\
               f"{self.exit} {self.kwargs}>"

    def __call__(self, inst):
        if self.ncalls is None:
            self.hook(inst, **self.kwargs)
            if self.exit:
                sys.exit()
        else:
            if self.ncalls > 0:
                self.ncalls -= 1
                self.hook(inst, **self.kwargs)
                if self.exit:
                    sys.exit()

    def copy(self):
        hm = _HookMeta(self.class_name, self.inst_id, self.hook, self.ncalls,
                       self.exit, **self.kwargs)
        # keep the same stamp so that the order of hooks doesn't change
        hm._stamp = self._stamp
        if self.inst_id is None:
            self.children.append(hm)

        return hm

    def deactivate(self):
        self.ncalls = 0
        for child in self.children:
            child.deactivate()


class _HookDecorator(object):
    def __init__(self, inst, func, hooks):
        self.__name__ = func.__name__
        self.__doc__ = func.__doc__

        self.inst = weakref.ref(inst)
        self.func = func
        self.pre_hooks = []
        self.post_hooks = []
        self.add_hooks(hooks)

    def add_hooks(self, hooks):
        """
        Add additional pre and/or post hooks to this method.

        Parameters
        ----------
        hooks : list
            List of hook data.
        """
        # copy any hooks with inst_id of None so we can modify them without modifying the original
        for prehook, posthook in hooks:
            if prehook is not None:
                if prehook.inst_id is None:
                    prehook = prehook.copy()
                self.pre_hooks.append(prehook)
            if posthook is not None:
                if posthook.inst_id is None:
                    posthook = posthook.copy()
                self.post_hooks.append(posthook)

        # put the hooks in order by stamp
        self.pre_hooks = sorted(self.pre_hooks, key=lambda x: x._stamp)
        self.post_hooks = sorted(self.post_hooks, key=lambda x: x._stamp)

    def _run_hooks(self, hooks):
        """
        Run the given list of hooks.

        Parameters
        ----------
        hooks : list
            List of hook data.
        inst : object
            Object instance to pass to hook functions.
        """
        inst = self.inst()
        if inst is None:
            return

        for hook in hooks:
            hook(inst)

    def __call__(self, *args, **kwargs):
        """
        Run the function with any pre and post hooks.

        Parameters
        ----------
        *args : list
            Positional arguments.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        object
            The return value of the function.
        """
        self._run_hooks(self.pre_hooks)
        ret = self.func(*args, **kwargs)
        self._run_hooks(self.post_hooks)
        return ret


def _setup_hooks(obj):
    """
    Wrap object's methods with a hook checking function if hooks are defined for any of them.

    Parameters
    ----------
    obj : object
        The object whose methods may be wrapped.
    """
    global _hooks

    # _setup_hooks should be called after 'obj' can return a valid name from _get_inst_id().
    # For example, in Problem, it can happen in __init__, but in Component and Group it shouldn't
    # happen until _setup_procs because that's the earliest point where the component/group has a
    # valid pathname.
    if use_hooks:

        # any object where we register hooks must define the '_get_inst_id' method.
        ident = obj._get_inst_id()

        if ident is None:
            raise RuntimeError(f"Object {obj} must define a '_get_inst_id' method that returns "
                               "a unique identifier for the object.")

        all_hooks = {}

        for c in getmro(obj.__class__):
            if c.__qualname__ in _hooks:
                classmeta = _hooks[c.__qualname__]

                if ident in classmeta:
                    instmeta = classmeta[ident]
                    for funcname, hooklist in instmeta.items():
                        method = getattr(obj, funcname, None)
                        if method is not None and callable(method):
                            if funcname not in all_hooks:
                                all_hooks[funcname] = []
                            all_hooks[funcname].extend(hooklist)

                if None in classmeta:
                    instmeta = classmeta[None]
                    for funcname, hooklist in instmeta.items():
                        method = getattr(obj, funcname, None)
                        if method is not None and callable(method):
                            if funcname not in all_hooks:
                                all_hooks[funcname] = []
                            all_hooks[funcname].extend(hooklist)

        for funcname, hooks in all_hooks.items():
            method = getattr(obj, funcname)
            if isinstance(method, _HookDecorator):
                method.add_hooks(hooks)
            else:
                setattr(obj, funcname, _HookDecorator(obj, method, hooks))


def _register_hook(fname, class_name, inst_id=None, pre=None, post=None, ncalls=None, exit=False,
                   **kwargs):
    """
    Register a hook function.

    Note that the 'class_name' arg should be the __qualname__ of the class, so for a nested class,
    the name would include the names of any containing classes as well, with each class name
    separated by a '.'.

    Parameters
    ----------
    fname : str
        The name of the function where the pre and/or post hook will be applied.
    class_name : str
        The name of the class owning the method where the hook will be applied.
    inst_id : str or None
        The name of the instance owning the method where the hook will be applied.
    pre : function (None)
        If not None, this hook will run before the named function runs.
    post : function (None)
        If not None, this hook will run after the named function runs.
    ncalls : int or None
        Auto-remove the hook function after this many calls.  If None, never auto-remove.
        If both pre and post are registered, this will affect both.
    exit : bool
        If True, run sys.exit() after calling the hook function.  If post is registered, this
        affects only post, else it will affect pre.
    **kwargs : dict of keyword arguments
        Keyword arguments that will be passed to the hook function.
    """
    if pre is None and post is None:
        raise RuntimeError("In _register_hook you must specify pre or post.")

    global _hooks
    if class_name in _hooks:
        cmeta = _hooks[class_name]
    else:
        _hooks[class_name] = cmeta = {}

    if inst_id in cmeta:
        imeta = cmeta[inst_id]
    else:
        cmeta[inst_id] = imeta = {}

    if fname not in imeta:
        imeta[fname] = []

    if pre is None:
        pre_hook = None
    else:
        pre_exit = exit if post is None else False
        pre_hook = _HookMeta(class_name, inst_id, pre, ncalls, pre_exit, **kwargs)

    if post is None:
        post_hook = None
    else:
        post_hook = _HookMeta(class_name, inst_id, post, ncalls, exit, **kwargs)

    imeta[fname].append((pre_hook, post_hook))


def _deactivate_hook(to_remove, hook):
    """
    Deactivate a hook function.

    Parameters
    ----------
    to_remove : bool or function
        If True, all hook functions in 'hooks' will be removed.  If a function, any function
        in 'hooks' that matches will be removed.
    hook : _HookMeta or None
        Hook metadata object.
    """
    if to_remove and hook is not None:
        if to_remove is True:
            hook.deactivate()
        else:  # to_remove is a hook function
            if to_remove is hook.hook:
                hook.deactivate()


def _unregister_hook(fname, class_name, inst_id=None, pre=True, post=True):
    """
    Unregister a hook function.

    By default, both pre and post hooks will be deactivated if they are present. To avoid
    removal of pre or post, you must set the pre or post arg to False.

    Parameters
    ----------
    fname : str
        The name of the function where the pre and/or post hook will be deactivated.
    class_name : str
        The name of the class owning the method where the hook will be deactivated.
    inst_id : str, optional
        The name of the instance owning the method where the hook will be deactivated.
    pre : bool or function, (True)
        If True, hooks that run before the named function runs will be deactivated. If pre is a
        function, then that function will have its hook(s) deactivated.
    post : bool or function, (True)
        If True, hooks that run after the named function runs will be deactivated. If post is a
        function, then that function will have its hook(s) deactivated.
    """
    try:
        classhooks = _hooks[class_name]
    except KeyError:
        return

    for instkey, hookdict in classhooks.items():
        if not (inst_id is None or instkey == inst_id):
            continue
        if fname in hookdict:
            for pre_hook, post_hook in hookdict[fname]:
                _deactivate_hook(pre, pre_hook)
                _deactivate_hook(post, post_hook)
        else:
            warnings.warn(f"No hook found for method '{fname}' for class '{class_name}' and "
                          f"instance '{inst_id}'.")



""" Unit tests for the system methods for modifying bounds and scalers."""
import unittest
from copy import deepcopy

import numpy as np

from openmdao.api import Problem, ExecComp, ExplicitComponent
from openmdao.components.linear_system_comp import LinearSystemComp
from openmdao.core.implicitcomponent import ImplicitComponent
from openmdao.core.indepvarcomp import IndepVarComp
from openmdao.drivers.scipy_optimizer import ScipyOptimizeDriver
from openmdao.solvers.linear.direct import DirectSolver
from openmdao.solvers.linear.linear_block_gs import LinearBlockGS
from openmdao.solvers.nonlinear.broyden import BroydenSolver
from openmdao.solvers.nonlinear.nonlinear_block_gs import NonlinearBlockGS
from openmdao.test_suite.components.expl_comp_array import TestExplCompArraySparse
from openmdao.test_suite.components.implicit_newton_linesearch import ImplCompTwoStates
from openmdao.test_suite.components.sellar import SellarDerivatives, SellarProblem
from openmdao.test_suite.components.paraboloid import Paraboloid
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.indexer import slicer


class TestSystemSetSolverDesvarConstraintObjectiveOutputOptions(unittest.TestCase):
    # this class is for tests that cut across all set_xxxxx_options methods
    def test_set_options_called_with_nonexistant_var(self):
        prob = Problem()
        prob.model.add_subsystem('comp', ImplCompTwoStatesNoMetadataSetInAddOutput())
        prob.model.set_output_solver_options(name='comp.bad_var', lower=1.5, upper=2.5)
        with self.assertRaises(RuntimeError) as ctx:
            prob.setup()
        self.assertEqual(str(ctx.exception),
                         "Output solver options set using System.set_output_solver_options for "
                         "non-existent variable 'comp.bad_var' in System ''.")

        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_design_var('x')
        prob.model.add_design_var('y')
        prob.model.add_constraint('con1')
        prob.model.add_constraint('con2')
        prob.model.add_objective('obj')
        with self.assertRaises(RuntimeError) as ctx:
            prob.model.set_design_var_options(name='bad_var', lower=-100, upper=100)
        self.assertEqual(str(ctx.exception),
                         "<class SellarDerivatives>: set_design_var_options called with design "
                         "variable 'bad_var' "
                         "that does not exist.")

        with self.assertRaises(RuntimeError) as ctx:
            prob.model.set_constraint_options(name='bad_var', ref0=-100.0, ref=100)
        self.assertEqual(str(ctx.exception),
                         "<class SellarDerivatives>: set_constraint_options called with "
                         "constraint variable 'bad_var' that does not exist.")

        with self.assertRaises(RuntimeError) as ctx:
            prob.model.set_objective_options(name='bad_var', ref0=-100.0, ref=100)
        self.assertEqual(str(ctx.exception),
                         "<class SellarDerivatives>: set_objective_options called with "
                         "objective variable 'bad_var' that does not exist.")


class TestSystemSetSolverOutputOptions(unittest.TestCase):
    # For tests specifically for set_output_solver_options
    # Some tests run models that are from other tests in which the
    #   the metadata is set using add_output arguments. Checking to see if we get same results
    #   when setting the metadata when using set_output_solver_options vs. add_output
    def _setup_model_for_lower_upper_tests(self, comp_class):
        prob = Problem()
        prob.model.add_subsystem('comp', comp_class(),
                                 promotes=['*'])
        prob.model.nonlinear_solver = BroydenSolver()
        prob.model.nonlinear_solver.options['maxiter'] = 25
        prob.model.nonlinear_solver.options['diverge_limit'] = 0.5
        prob.model.nonlinear_solver.options['state_vars'] = ['y', 'z']
        prob.model.linear_solver = DirectSolver()
        return prob

    def test_set_output_solver_options_top_model(self):
        # No metadata set on comp.z in the model. Set it after the fact. See if the results of
        #   running the model are as expected
        prob = self._setup_model_for_lower_upper_tests(ImplCompTwoStatesNoMetadataSetInAddOutput)
        prob.model.set_output_solver_options(name='comp.z', lower=1.5, upper=2.5)
        prob.setup()

        comp_z_meta = prob.model._var_allprocs_abs2meta['output']['comp.z']
        self.assertEqual(comp_z_meta['lower'], 1.5)
        self.assertEqual(comp_z_meta['upper'], 2.5)

        # Test lower bound: should go to the lower bound and stall
        prob['comp.x'] = 2.0
        prob['comp.y'] = 0.0
        prob['comp.z'] = 1.6
        prob.run_model()
        assert_near_equal(prob['comp.z'], [1.5], 1e-8)

        # Test upper bound: should go to the upper bound and stall
        prob['comp.x'] = 0.5
        prob['comp.y'] = 0.0
        prob['comp.z'] = 2.4
        prob.run_model()
        assert_near_equal(prob['comp.z'], [2.5], 1e-8)

    def test_set_output_solver_options_sub_system(self):
        # Try setting the metadata using set_output_solver_options on a subsystem, not the top
        prob = self._setup_model_for_lower_upper_tests(ImplCompTwoStatesNoMetadataSetInAddOutput)
        comp = prob.model._get_subsystem('comp')
        comp.set_output_solver_options(name='z', lower=1.5, upper=2.5)
        prob.setup()

        comp_z_meta = prob.model._var_allprocs_abs2meta['output']['comp.z']
        self.assertEqual(comp_z_meta['lower'], 1.5)
        self.assertEqual(comp_z_meta['upper'], 2.5)

        # Test lower bound: should go to the lower bound and stall
        prob['comp.x'] = 2.0
        prob['comp.y'] = 0.0
        prob['comp.z'] = 1.6
        prob.run_model()
        assert_near_equal(prob['comp.z'], [1.5], 1e-8)

        # Test upper bound: should go to the upper bound and stall
        prob['comp.x'] = 0.5
        prob['comp.y'] = 0.0
        prob['comp.z'] = 2.4
        prob.run_model()
        assert_near_equal(prob['comp.z'], [2.5], 1e-8)

    def test_set_output_solver_options_set_lower_upper_to_none(self):
        # test the ability to override values set with add_output
        #  In this case, set the lower and upper to None, to see if the code handles that
        prob = self._setup_model_for_lower_upper_tests(ImplCompTwoStates)
        prob.model.set_output_solver_options(name='comp.z', lower=None, upper=None)
        prob.setup()

        comp_z_meta = prob.model._var_allprocs_abs2meta['output']['comp.z']
        self.assertEqual(comp_z_meta['lower'], None)
        self.assertEqual(comp_z_meta['upper'], None)

        prob.set_solver_print(level=0)
        # Test lower bound: should go to the lower bound and stall
        prob['comp.x'] = 2.0
        prob['comp.y'] = 0.0
        prob['comp.z'] = 1.6
        prob.run_model()
        assert_near_equal(prob['comp.z'], [1.3333333333333333], 1e-8)

    def test_set_output_solver_options_ref_no_meta(self):
        # Test setting scaling using set_output_solver_options
        prob = Problem()
        model = prob.model
        model.add_subsystem('comp', ScalingExample3NoMetadataSetInAddOutput(),
                            promotes_inputs=['x1', 'x2'])
        prob.model.set_input_defaults('x1', 1.0)
        prob.model.set_input_defaults('x2', 1.0)
        model.set_output_solver_options(name='comp.y1', ref=1e2, res_ref=1e5)
        model.set_output_solver_options(name='comp.y2', ref=1e3, res_ref=1e-5)
        prob.setup()

        # See if values were set in the metadata
        comp_y1_meta = model._var_allprocs_abs2meta['output']['comp.y1']
        self.assertEqual(comp_y1_meta['ref'], 1e2)
        self.assertEqual(comp_y1_meta['res_ref'], 1e5)

        prob.run_model()
        model.run_apply_nonlinear()

        # are the results as expected with those values of ref and res_ref
        with model._scaled_context_all():
            val = model.comp._residuals['y1']
            assert_near_equal(val, [-.995])
            val = model.comp._residuals['y2']
            assert_near_equal(val, [(1 - 6000.) / 6000.])

    def test_set_output_solver_affect_on_has_scaling_and_bounds_attributes(self):
        # make sure bounds and scaling attributes get set correctly with different
        #  combinations of setting via add_output and set_output_solver_options

        # ref and lower not set either in add_output of set_output_solver_options
        prob = Problem()
        comp = prob.model.add_subsystem('comp', ScalingExample3ForTesting())
        prob.setup()
        self.assertFalse(comp._has_bounds)
        self.assertFalse(comp._has_output_scaling)

        # ref and lower set with set_output_solver_options
        prob = Problem()
        comp = prob.model.add_subsystem('comp', ScalingExample3ForTesting())
        prob.model.set_output_solver_options(name='comp.y1', ref=1.2, lower=2)
        prob.setup()
        self.assertTrue(comp._has_bounds)
        self.assertTrue(comp._has_output_scaling)

        # ref and lower set with add_output on both variables
        #  Unset with set_output_solver_options on just y1
        prob = Problem()
        comp = prob.model.add_subsystem('comp', ScalingExample3ForTesting(ref1=1.2, lower1=2.0,
                                                                          ref2=1.2, lower2=2.0))
        prob.model.set_output_solver_options(name='comp.y1', ref=1.0, lower=None)
        prob.setup()
        self.assertTrue(comp._has_bounds)
        self.assertTrue(comp._has_output_scaling)

        # ref and lower set with add_output on both variables
        #  Unset with set_output_solver_options on both y1 and y2
        prob = Problem()
        comp = prob.model.add_subsystem('comp', ScalingExample3ForTesting(ref1=1.2, lower1=2.0,
                                                                          ref2=1.2, lower2=2.0))
        prob.model.set_output_solver_options(name='comp.y1', ref=1.0, lower=None)
        prob.model.set_output_solver_options(name='comp.y2', ref=1.0, lower=None)
        prob.setup()
        self.assertFalse(comp._has_bounds)
        self.assertFalse(comp._has_output_scaling)

    def test_set_output_solver_options_after_initial_run(self):
        # Testing to see if the error message occurs when the user runs a model
        #   and then calls set_output_solver_options, then tries to run model again
        #   without first running setup again
        prob = Problem()
        prob.model.add_subsystem('comp', ScalingExample3())
        prob.model.set_output_solver_options(name='comp.y1', ref=3)
        prob.setup()
        prob.run_model()
        prob.model.set_output_solver_options(name='comp.y1', ref=4)
        msg = "Problem .*: Before calling `run_model`, the `setup` method must be called if " \
              "set_output_solver_options has been called."
        with self.assertRaisesRegex(RuntimeError, msg) as cm:
            prob.run_model()

        # also make sure it doesn't raise an error if done correctly with an additional call
        #  to setup
        prob = Problem()
        prob.model.add_subsystem('comp', ScalingExample3())
        prob.model.set_output_solver_options(name='comp.y1', ref=3)
        prob.setup()
        prob.run_model()
        prob.model.set_output_solver_options(name='comp.y1', ref=4)
        prob.setup()
        prob.run_model()

    def test_set_output_solver_options_override_all_options(self):
        # Just another test to check setting all metadata in one call
        prob = Problem()
        model = prob.model
        # ScalingExample3 sets values for ref, res_ref, ref0, lower, and upper
        comp = model.add_subsystem('comp', ScalingExample3())

        # override all those values
        model.set_output_solver_options(name='comp.y1',
                                        ref=3, res_ref=6, ref0=9, lower=12, upper=15)
        prob.setup()

        # check that they were all set
        comp_y1_meta = model._var_allprocs_abs2meta['output']['comp.y1']
        self.assertEqual(comp_y1_meta['ref'], 3)
        self.assertEqual(comp_y1_meta['res_ref'], 6)
        self.assertEqual(comp_y1_meta['ref0'], 9)
        self.assertEqual(comp_y1_meta['lower'], 12)
        self.assertEqual(comp_y1_meta['upper'], 15)

    def test_set_output_solver_options_no_setup_after_call_to_set_output_solver_options(self):
        prob = SellarProblem(nonlinear_solver=NonlinearBlockGS,
                             linear_solver=LinearBlockGS)
        lscomp = prob.model.add_subsystem('lscomp', LinearSystemComp())
        lscomp.nonlinear_solver = NonlinearBlockGS(maxiter=5)
        prob.model.set_output_solver_options(name='lscomp.x', ref=3)
        prob.setup()
        prob.run_driver()

        prob.model.set_output_solver_options(name='lscomp.x', ref=5)
        msg = "Problem .*: Before calling `run_driver`, the `setup` method must be called if " \
              "set_output_solver_options has been called."
        with self.assertRaisesRegex(RuntimeError, msg) as cm:
            prob.run_driver()

    def test_set_output_solver_options_vector_values(self):
        class SimpleImpComp(ImplicitComponent):
            def setup(self):
                self.add_output('y', np.array([3, 1]), upper = np.array([10,10]))
                self.declare_partials(of='*', wrt='*')
        prob = Problem()
        prob.model.add_subsystem('px', SimpleImpComp())
        prob.setup()

        new_upper_solver_option = np.array([1,1])
        prob.model.set_output_solver_options(name='px.y', upper = new_upper_solver_option)
        prob.setup()
        output_meta = prob.model._var_allprocs_abs2meta['output']
        assert_near_equal(output_meta['px.y']['upper'], new_upper_solver_option)

class TestSystemSetDesignVarOptions(unittest.TestCase):
    def test_set_design_var_options_scaler_adder_ref_ref0(self):
        # See if setting options for des vars results in the same metadata whether
        #   they are set using add_design_var or set_design_var_options

        # first set the options ref and ref0 using add_design_var
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_design_var('x', lower=-100, upper=100, ref0=-100.0, ref=100)
        prob.model.add_design_var('z', lower=-100, upper=100)
        prob.setup()
        des_vars_using_add_design_var = prob.model.get_design_vars()

        # then set the options using set_design_var_options
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_design_var('x')
        prob.model.add_design_var('z')
        prob.model.set_design_var_options(name='x', lower=-100, upper=100, ref0=-100.0, ref=100)
        prob.model.set_design_var_options(name='z', lower=-100, upper=100)
        prob.setup()
        des_vars_using_set_design_var_options = prob.model.get_design_vars()

        self.assertEqual(des_vars_using_add_design_var, des_vars_using_set_design_var_options)

        # Now do the same using scaler and adder
        # first set the options using add_design_var
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_design_var('x', lower=-100, upper=100, adder=3, scaler=10)
        prob.model.add_design_var('z', lower=-100, upper=100)
        prob.setup()
        des_vars_using_add_design_var = prob.model.get_design_vars()

        # then set the options using set_design_var_options
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_design_var('x')
        prob.model.add_design_var('z')
        prob.model.set_design_var_options(name='x', lower=-100, upper=100, adder=3, scaler=10)
        prob.model.set_design_var_options(name='z', lower=-100, upper=100)
        prob.setup()
        des_vars_using_set_design_var_options = prob.model.get_design_vars()

        self.assertEqual(des_vars_using_add_design_var, des_vars_using_set_design_var_options)

    def test_set_design_var_options_adder_scalar_ref_ref0_override(self):
        # first set the options for scaler and adder using add_design_var and then
        #   override with calls to set_design_var_options. Compare
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_design_var('x', lower=-100, upper=100, scaler=3.5, adder=77.0)
        prob.model.add_design_var('z', lower=-100, upper=100)
        prob.setup()

        des_vars_using_add_design_var = prob.model.get_design_vars()

        # then set the options for scaler and adder using set_design_var_options
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_design_var('x', lower=-100, upper=100, ref0=-2.0, ref=20)
        prob.model.add_design_var('z', lower=-100, upper=100)
        prob.model.set_design_var_options(name='x', scaler=3.5, adder=77.0)
        prob.setup()
        des_vars_using_set_design_var_options = prob.model.get_design_vars()

        assert_near_equal(des_vars_using_add_design_var, des_vars_using_set_design_var_options)

        # Now set the options for ref and ref0 using add_design_var
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_design_var('x', lower=-100, upper=100, ref0=-2.0, ref=20)
        prob.model.add_design_var('z', lower=-100, upper=100)
        prob.setup()
        des_vars_using_add_design_var = prob.model.get_design_vars()

        # then set the options for ref and ref0 using set_design_var_options
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_design_var('x', lower=-100, upper=100, scaler=3.5, adder=77.0)
        prob.model.add_design_var('z', lower=-100, upper=100)
        prob.model.set_design_var_options(name='x', ref0=-2.0, ref=20)
        prob.setup()
        des_vars_using_set_design_var_options = prob.model.get_design_vars()

        assert_near_equal(des_vars_using_add_design_var, des_vars_using_set_design_var_options)

    def test_set_design_var_options_post_setup_run_driver_compare_outputs(self):
        # Check to make sure that set_design_var_options can be called after setup
        #   and have the same effect as if setting the metadata were part of add_design_var
        prob = Problem()
        model = prob.model
        model.add_subsystem('comp', Paraboloid(), promotes=['*'])
        model.add_subsystem('con', ExecComp('c = x - y'), promotes=['*'])
        prob.set_solver_print(level=0)
        prob.driver = ScipyOptimizeDriver(optimizer='SLSQP', tol=1e-9, disp=False)
        model.add_design_var('x', lower=-50.0, upper=50.0)
        model.add_design_var('y', lower=0.0, upper=50.0)
        model.add_objective('f_xy')
        model.add_constraint('c', lower=15.0, linear=True)
        model.set_input_defaults('x', val=0.0)
        model.set_input_defaults('y', val=0.0)
        prob.setup()
        prob.run_driver()
        x_when_setting_using_add_design_var = prob['x']
        y_when_setting_using_add_design_var = prob['y']

        # now do the same but set the constraint using set_design_var_options
        prob = Problem()
        model = prob.model
        model.add_subsystem('comp', Paraboloid(), promotes=['*'])
        model.add_subsystem('con', ExecComp('c = x - y'), promotes=['*'])
        prob.set_solver_print(level=0)
        prob.driver = ScipyOptimizeDriver(optimizer='SLSQP', tol=1e-9, disp=False)
        model.add_design_var('x')
        model.add_design_var('y')
        model.add_objective('f_xy')
        model.add_constraint('c', lower=15.0, linear=True)
        model.set_input_defaults('x', val=0.0)
        model.set_input_defaults('y', val=0.0)
        prob.setup()
        model.set_design_var_options('x',lower=-50.0, upper=50.0 )
        model.set_design_var_options('y',lower= 0.0, upper=50.0 )
        prob.run_driver()
        assert_near_equal(prob['x'], x_when_setting_using_add_design_var, 1e-6)
        assert_near_equal(prob['y'], y_when_setting_using_add_design_var, 1e-6)

    def test_set_design_var_options_vector_values(self):
        prob = Problem(model=SellarDerivatives())
        model = prob.model
        model.nonlinear_solver = NonlinearBlockGS()

        prob.driver = ScipyOptimizeDriver()
        prob.driver.options['optimizer'] = 'SLSQP'
        prob.driver.options['tol'] = 1e-9

        model.add_design_var('z', lower=np.array([-10.0, 0.0]), upper=np.array([10.0, 10.0]))
        model.add_design_var('x', lower=0.0, upper=10.0)
        model.add_objective('obj')
        model.add_constraint('con1', lower=-100, upper=100)
        model.add_constraint('con2', upper=0.0)

        prob.setup()
        new_z_lower_bound = np.array([0.0, 0.0])
        model.set_design_var_options("z", lower=np.array(new_z_lower_bound))
        des_vars_using_set_design_var_options = prob.model.get_design_vars()
        assert_near_equal(des_vars_using_set_design_var_options['z']['lower'], new_z_lower_bound)


class TestSystemSetConstraintsOptions(unittest.TestCase):

    def test_set_constraints_options_lower_upper(self):
        # first set the options lower and upper using add_constraint
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_constraint('con1', lower=-100, upper=100)
        prob.model.add_constraint('con2')
        prob.setup()
        constraints_using_add_constraint = prob.model.get_constraints()

        # then set the options using set_constraint_options
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_constraint('con1')
        prob.model.add_constraint('con2')
        prob.model.set_constraint_options(name='con1', lower=-100, upper=100)
        prob.setup()
        constraints_using_set_objective_options = prob.model.get_constraints()

        self.assertEqual(constraints_using_add_constraint, constraints_using_set_objective_options)

    def test_set_constraints_options_equals_overrides_lower_upper(self):
        # First set lower and upper in add_constraint
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_constraint('con1', equals=345.)
        prob.model.add_constraint('con2')
        prob.setup()
        constraints_using_add_constraint = prob.model.get_constraints()

        # Then set am equals constraint using an override via set_constraint_options
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_constraint('con1', lower=-100, upper=100)
        prob.model.add_constraint('con2')
        prob.model.set_constraint_options('con1', equals=345.)
        prob.setup()
        constraints_using_set_constraint_options = prob.model.get_constraints()
        self.assertEqual(constraints_using_add_constraint, constraints_using_set_constraint_options)

    def test_set_constraints_options_lower_upper_overrides_equals(self):
        # First set lower and upper in add_constraint
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_constraint('con1', lower=-100, upper=100)
        prob.model.add_constraint('con2')
        prob.setup()
        constraints_using_add_constraint = prob.model.get_constraints()

        # Then set it using an override via set_constraint_options
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_constraint('con1', equals=345.)
        prob.model.add_constraint('con2')
        prob.model.set_constraint_options('con1', lower=-100, upper=100)
        prob.setup()
        constraints_using_set_constraint_options = prob.model.get_constraints()

        self.assertEqual(constraints_using_add_constraint, constraints_using_set_constraint_options)

    def test_set_constraints_options_scaler_adder_ref_ref0(self):
        # first set the options ref and ref0 using add_constraint
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_constraint('con1', ref0=-100.0, ref=100)
        prob.model.add_constraint('con2')
        prob.setup()
        constraints_using_add_constraint = prob.model.get_constraints()

        # then set the options using set_constraint_options
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_constraint('con1')
        prob.model.add_constraint('con2')
        prob.model.set_constraint_options(name='con1', ref0=-100.0, ref=100)
        prob.setup()
        constraints_using_set_constraint_options = prob.model.get_constraints()

        self.assertEqual(constraints_using_add_constraint, constraints_using_set_constraint_options)

        # Now do the same using scaler and adder
        # first set the options using add_constraint
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_constraint('con1', adder=3, scaler=10)
        prob.model.add_constraint('con2')
        prob.setup()
        constraints_using_add_constraint = prob.model.get_constraints()

        # then set the options using set_constraint_options
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_constraint('con1')
        prob.model.add_constraint('con2')
        prob.model.set_constraint_options(name='con1', adder=3, scaler=10)
        prob.setup()
        constraints_using_set_constraint_options = prob.model.get_constraints()

        self.assertEqual(constraints_using_add_constraint, constraints_using_set_constraint_options)

    def test_set_constraints_options_adder_scalar_ref_ref0_override(self):
        # Test overriding ref values set in add_constraint with scaler and adder
        #     values set in set_constraint_options
        # Then do the opposite

        # first set the options for scaler and adder using add_constraint
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_constraint('con1', scaler=3.5, adder=77.0, lower=-100, upper=100)
        prob.model.add_constraint('con2')
        prob.setup()
        constraints_using_add_constraint = prob.model.get_constraints()

        # then set the options for scaler and adder using set_constraint_options
        #  overriding the ref0 and ref set in add_constraint
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_constraint('con1', ref0=-2.0, ref=20, lower=-100, upper=100)
        prob.model.add_constraint('con2')
        prob.model.set_constraint_options(name='con1', scaler=3.5, adder=77.0)
        prob.setup()
        constraints_using_set_constraint_options = prob.model.get_constraints()

        assert_near_equal(constraints_using_add_constraint,
                          constraints_using_set_constraint_options)

        # first set the options for ref and ref0 using add_constraint
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_constraint('con1', ref0=-2.0, ref=20, lower=-100, upper=100)
        prob.model.add_constraint('con2')
        prob.setup()
        constraints_using_add_constraint = prob.model.get_constraints()

        # then set the options for ref and ref0 using set_constraint_options
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_constraint('con1', scaler=3.5, adder=77.0, lower=-100, upper=100)
        prob.model.add_constraint('con2')
        prob.model.set_constraint_options('con1', ref0=-2.0, ref=20)
        prob.setup()
        constraints_using_set_constraint_options = prob.model.get_constraints()

        assert_near_equal(constraints_using_add_constraint,
                          constraints_using_set_constraint_options)

    def test_set_constraint_options_with_aliases(self):
        # One complication is the path to constraints in set_constraint_options, since a single
        #  variable may have multiple constraints imposed upon it provided they are aliased,
        #  but we should be able to account for this. If there is ambiguity (an alias exists),
        #  raise an error.
        prob = Problem()
        exec_comp = prob.model.add_subsystem('exec', ExecComp())
        exec_comp.add_expr('y = x**2', y={'shape': (10, 3)}, x={'shape': (10, 3)})
        prob.model.add_constraint('exec.y', alias='y0', indices=slicer[0, -1], scaler=3.5,
                                  adder=77.0)
        prob.model.add_constraint('exec.y', alias='yf', indices=slicer[-1, 0], equals=0)
        prob.setup()
        constraints_using_add_constraint = prob.model.get_constraints()

        prob = Problem()
        exec_comp = prob.model.add_subsystem('exec', ExecComp())
        exec_comp.add_expr('y = x**2', y={'shape': (10, 3)}, x={'shape': (10, 3)})
        prob.model.add_constraint('exec.y', alias='y0', indices=slicer[0, -1])
        prob.model.add_constraint('exec.y', alias='yf', indices=slicer[-1, 0], equals=0)
        prob.model.set_constraint_options('exec.y', alias='y0', scaler=3.5, adder=77.0)
        prob.setup()

        constraints_using_set_constraint_options = prob.model.get_constraints()
        assert_near_equal(constraints_using_add_constraint,
                          constraints_using_set_constraint_options)

    def test_set_constraint_options_error_calling_without_aliases(self):
        prob = Problem()
        exec_comp = prob.model.add_subsystem('exec', ExecComp())
        exec_comp.add_expr('y = x**2', y={'shape': (10, 3)}, x={'shape': (10, 3)})
        prob.model.add_constraint('exec.y', alias='y0', indices=slicer[0, -1], scaler=3.5,
                                  adder=77.0)
        prob.model.add_constraint('exec.y', alias='yf', indices=slicer[-1, 0], equals=0)

        with self.assertRaises(RuntimeError) as cm:
            prob.model.set_constraint_options('exec.y', ref0=-2.0, ref=20)
        msg = "<class Group>: set_constraint_options called with constraint variable 'exec.y' " \
              "that has multiple aliases: ['y0', 'yf']. Call set_objective_options with the " \
              "'alias' argument set to one of those aliases."
        self.assertEqual(str(cm.exception), msg)

    def test_set_constraint_options_vector_values(self):
        prob = Problem()
        model = prob.model
        model.add_subsystem('p1', IndepVarComp('widths', np.zeros((2, 2))), promotes=['*'])
        model.add_subsystem('comp', TestExplCompArraySparse(), promotes=['*'])
        model.add_subsystem('obj',
                            ExecComp('o = areas[0, 0] + areas[1, 1]', areas=np.zeros((2, 2))),
                            promotes=['*'])

        prob.set_solver_print(level=0)

        prob.driver = ScipyOptimizeDriver(optimizer='SLSQP', tol=1e-9, disp=False)

        model.add_design_var('widths', lower=-50.0, upper=50.0)
        model.add_objective('o')
        model.add_constraint('areas', equals=np.array([24.0, 21.0, 3.5, 17.5]))

        prob.setup()

        new_areas_equals_bound = np.array([4.0, 1.0, 0.5, 7.5])
        prob.model.set_constraint_options(name='areas', equals=new_areas_equals_bound)
        constraints_using_set_objective_options = prob.model.get_constraints()
        assert_near_equal(constraints_using_set_objective_options['areas']['equals'],
                          new_areas_equals_bound)


class TestSystemSetObjectiveOptions(unittest.TestCase):
    def test_set_objective_options_scaler_adder_ref_ref0(self):
        # See if setting options for objective results in the same metadata whether
        #   they are set using add_objective or set_objective_options

        # first set the options ref and ref0 using add_objective
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_objective('obj', ref0=-100.0, ref=100)
        prob.setup()
        objective_using_add_objective = prob.model.get_objectives()

        # then set the options using set_objective_options
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_objective('obj')
        prob.model.set_objective_options(name='obj', ref0=-100.0, ref=100)
        prob.setup()
        objective_using_set_objective_options = prob.model.get_objectives()

        self.assertEqual(objective_using_add_objective, objective_using_set_objective_options)

        # now do the same using scaler and adder
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_objective('obj', adder=3, scaler=10)
        prob.setup()
        objective_using_add_objective = prob.model.get_objectives()

        # then set the options using set_objective_options
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_objective('obj')
        prob.model.set_objective_options(name='obj', adder=3, scaler=10)
        prob.setup()
        objective_using_set_objective_options = prob.model.get_objectives()

        self.assertEqual(objective_using_add_objective, objective_using_set_objective_options)

    def test_set_objective_options_adder_scalar_ref_ref0_override(self):
        # first set the options for scaler and adder using add_objective
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_objective('obj', scaler=3.5, adder=77.0)
        prob.setup()
        objective_using_add_objective = prob.model.get_objectives()

        # then set the options for ref and ref0 using set_objective_options
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_objective('obj', ref0=-2.0, ref=20)
        prob.model.set_objective_options(name='obj', scaler=3.5, adder=77.0)
        prob.setup()
        objective_using_set_objective_options = prob.model.get_objectives()

        assert_near_equal(objective_using_add_objective, objective_using_set_objective_options)

        # first set the options for ref and ref0 using add_objective
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_objective('obj', ref0=-2.0, ref=20)
        prob.setup()
        objective_using_add_objective = prob.model.get_objectives()

        # then set the options for ref and ref0 using set_design_var_options
        prob = Problem()
        prob.model = SellarDerivatives()
        prob.model.add_objective('obj', scaler=3.5, adder=77.0)
        prob.model.set_objective_options(name='obj', ref0=-2.0, ref=20)
        prob.setup()
        objective_using_set_objective_options = prob.model.get_objectives()

        assert_near_equal(objective_using_add_objective, objective_using_set_objective_options)

    def test_set_objective_options_with_alias(self):
        # Check to make sure setting options on the objective, where the objective is an alias,
        # works.
        # Testing both ref and ref0 plus scaler and adder
        # Optimize a problem with both using add_objective to set the options and also
        #   by using set_objective_options. Compare the optimization results and also the
        #   objective metadata

        def _build_model():
            prob = Problem()
            model = prob.model
            model.add_subsystem('comp', ParaboloidObjConsArray(), promotes=['*'])
            prob.driver = ScipyOptimizeDriver(optimizer='SLSQP', tol=1e-9, disp=False)
            prob.set_solver_print(level=0)
            model.add_design_var('x', lower=-50.0, upper=50.0)
            model.add_design_var('y', lower=-50.0, upper=50.0)
            return prob

        # Set the ref and ref0 using add_objective. Save the output values
        #   to compare results when using set_objective_options
        prob = _build_model()
        prob.model.add_constraint('fg_xy', indices=[1], alias='g_xy', lower=0, upper=10)
        prob.model.add_objective('fg_xy', index=0, alias='f_xy', ref=3.0, ref0=2.0)
        prob.setup()
        objective_using_add_objective = deepcopy(prob.model.get_objectives())
        prob.run_driver()

        fg_xy_using_add_objective = prob['fg_xy'][0]
        comp_x_using_add_objective = prob['comp.x']
        comp_y_using_add_objective = prob['comp.y']

        # Now use set_objective_options to set ref and ref0 and compare
        prob = _build_model()
        prob.model.add_constraint('fg_xy', indices=[1], alias='g_xy', lower=0, upper=10)
        prob.model.add_objective('fg_xy', index=0, alias='f_xy')
        prob.model.set_objective_options(name='fg_xy', alias='f_xy', ref=3.0, ref0=2.0)
        prob.setup()

        objective_using_set_objective_options = deepcopy(prob.model.get_objectives())
        assert_near_equal(objective_using_set_objective_options, objective_using_add_objective)

        prob.run_driver()
        assert_near_equal(prob['fg_xy'][0], fg_xy_using_add_objective, 1e-6)
        assert_near_equal(prob['comp.x'], comp_x_using_add_objective, 1e-4)
        assert_near_equal(prob['comp.y'], comp_y_using_add_objective, 1e-4)

        # Set the scaler and adder using add_objective. Save the output values
        #   to compare results when using set_objective_options
        prob = _build_model()
        prob.model.add_constraint('fg_xy', indices=[1], alias='g_xy', lower=0, upper=10)
        prob.model.add_objective('fg_xy', index=0, alias='f_xy', scaler=3.0, adder=2.0)
        prob.setup()
        objective_using_add_objective = deepcopy(prob.model.get_objectives())

        prob.run_driver()
        fg_xy_using_add_objective = prob['fg_xy'][0]
        comp_x_using_add_objective = prob['comp.x']
        comp_y_using_add_objective = prob['comp.y']

        # Now use set_objective_options to set scaler and adder and compare
        prob = _build_model()
        prob.model.add_constraint('fg_xy', indices=[1], alias='g_xy', lower=0, upper=10)
        prob.model.add_objective('fg_xy', index=0, alias='f_xy')
        prob.model.set_objective_options(name='fg_xy', alias='f_xy', scaler=3.0, adder=2.0)
        prob.setup()

        objective_using_set_objective_options = deepcopy(prob.model.get_objectives())
        assert_near_equal(objective_using_set_objective_options, objective_using_add_objective)

        prob.run_driver()
        assert_near_equal(prob['fg_xy'][0], fg_xy_using_add_objective, 1e-6)
        assert_near_equal(prob['comp.x'], comp_x_using_add_objective, 1e-4)
        assert_near_equal(prob['comp.y'], comp_y_using_add_objective, 1e-4)

    def test_set_objective_options_without_using_alias(self):
        prob = Problem()
        model = prob.model
        model.add_subsystem('comp', ParaboloidObjConsArray(), promotes=['*'])
        prob.driver = ScipyOptimizeDriver(optimizer='SLSQP', tol=1e-9, disp=False)
        prob.set_solver_print(level=0)
        prob.model.add_constraint('fg_xy', indices=[1], alias='g_xy', lower=0, upper=10)
        prob.model.add_objective('fg_xy', index=0, alias='f_xy')

        with self.assertRaises(RuntimeError) as cm:
            prob.model.set_objective_options(name='fg_xy', ref=3.0, ref0=2.0)
        msg = "<class Group>: set_objective_options called with objective variable 'fg_xy' that " \
              "has multiple aliases: ['g_xy', 'f_xy']. Call set_objective_options with the " \
              "'alias' argument set to one of those aliases."
        self.assertEqual(str(cm.exception), msg)


class ImplCompTwoStatesNoMetadataSetInAddOutput(ImplicitComponent):
    """
    A Simple Implicit Component with an additional output equation.

    This version has no metadata set in the calls to add_ouput
    """
    def setup(self):
        self.add_input('x', 0.5)
        self.add_output('y', 0.0)
        self.add_output('z', 2.0)

        self.maxiter = 10
        self.atol = 1.0e-12

    def setup_partials(self):
        self.declare_partials(of='*', wrt='*')

    def apply_nonlinear(self, inputs, outputs, residuals):
        x = inputs['x']
        y = outputs['y']
        z = outputs['z']

        residuals['y'] = y - x - 2.0 * z
        residuals['z'] = x * z + z - 4.0

    def linearize(self, inputs, outputs, jac):
        jac[('y', 'x')] = -1.0
        jac[('y', 'y')] = 1.0
        jac[('y', 'z')] = -2.0

        jac[('z', 'z')] = inputs['x'] + 1.0
        jac[('z', 'x')] = outputs['z']


class ScalingExample3(ImplicitComponent):
    def setup(self):
        self.add_input('x1', val=100.0)
        self.add_input('x2', val=5000.0)
        self.add_output('y1', val=200., ref=1e2, res_ref=1e5, ref0=1.0, lower=2., upper=3)
        self.add_output('y2', val=6000., ref=1e3, res_ref=1e-5)

    def apply_nonlinear(self, inputs, outputs, residuals):
        x1 = inputs['x1']
        x2 = inputs['x2']
        y1 = outputs['y1']
        y2 = outputs['y2']

        residuals['y1'] = 1e5 * (x1 - y1) / y1
        residuals['y2'] = 1e-5 * (x2 - y2) / y2


class ScalingExample3NoMetadataSetInAddOutput(ImplicitComponent):
    def setup(self):
        self.add_input('x1', val=100.0)
        self.add_input('x2', val=5000.0)
        self.add_output('y1', val=200.)
        self.add_output('y2', val=6000.)

    def apply_nonlinear(self, inputs, outputs, residuals):
        x1 = inputs['x1']
        x2 = inputs['x2']
        y1 = outputs['y1']
        y2 = outputs['y2']

        residuals['y1'] = 1e5 * (x1 - y1) / y1
        residuals['y2'] = 1e-5 * (x2 - y2) / y2

class ScalingExample3ForTesting(ImplicitComponent):
    def __init__(self, ref1=1.0, lower1=None, ref2=1.0, lower2=None):
        super().__init__()
        self.ref1 = ref1
        self.lower1 = lower1
        self.ref2 = ref2
        self.lower2 = lower2

    def setup(self):
        self.add_input('x1', val=100.0)
        self.add_input('x2', val=5000.0)
        self.add_output('y1', val=200., ref=self.ref1, lower=self.lower1)
        self.add_output('y2', val=200., ref=self.ref2, lower=self.lower2)

    def apply_nonlinear(self, inputs, outputs, residuals):
        x1 = inputs['x1']
        x2 = inputs['x2']
        y1 = outputs['y1']
        y2 = outputs['y2']
        residuals['y1'] = 1e5 * (x1 - y1) / y1
        residuals['y2'] = 1e-5 * (x2 - y2) / y2

class ParaboloidObjConsArray(ExplicitComponent):
    def setup(self):
        self.add_input('x', val=0.0)
        self.add_input('y', val=0.0)
        self.add_output('fg_xy', val=np.array([0., 0.]))

    def setup_partials(self):
        self.declare_partials('*', '*')

    def compute_partials(self, inputs, partials):
        x = inputs['x']
        y = inputs['y']

        partials['fg_xy', 'x'][0] = 2 * (x - 3) + y
        partials['fg_xy', 'x'][1] = 1.0

        partials['fg_xy', 'y'][0] = x + 2 * (y + 4)
        partials['fg_xy', 'y'][1] = 1.0

    def compute(self, inputs, outputs):
        x = inputs['x']
        y = inputs['y']
        outputs['fg_xy'] = [
            (x - 3.0) ** 2 + x * y + (y + 4.0) ** 2 - 3.0,
            x + y
        ]

if __name__ == "__main__":
    unittest.main()
